# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""This is the main module that defines what a QComponent is in Qiskit Metal.

To see the docstring of QComponent in Jupyter notebook, use:
>> ?QComponent
"""
# pylint: disable=too-many-lines, too-many-public-methods

import logging
import inspect
import random
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Iterable, List, Union, Tuple, Dict as Dict_
from datetime import datetime
import pandas as pd
import numpy as np
import pprint
from inspect import signature

from qiskit_metal import draw
from qiskit_metal import is_design, logger

import qiskit_metal.qlibrary as qlibrary
from qiskit_metal import config
from qiskit_metal.draw import BaseGeometry
from qiskit_metal.toolbox_python.attr_dict import Dict
from qiskit_metal.toolbox_python.display import format_dict_ala_z
from qiskit_metal.qlibrary.core._parsed_dynamic_attrs import ParsedDynamicAttributes_Component

if not config.is_building_docs():
    from ...draw import Vector

__all__ = ['QComponent']

if TYPE_CHECKING:
    # For linting typechecking, import modules that can't be loaded here under normal conditions.
    # For example, I can't import QDesign, because it requires QComponent first. We have the
    # chicken and egg issue.
    from ...designs import QDesign
    import matplotlib


class QComponent():
    """`QComponent` is the core class for all Metal components and is the
    central construct from which all components in Metal are derived.

    The class defines the user interface for working with components.

    For front-end user:
        * Manipulates the dictionary of options (stored as string-string key-value
          pairs) to change the geometry and properties of the component.
        * The options of the class are stored in an options dictionary. These
          include the geometric sizes, such as width='10um' or height='5mm', etc.
        * The `make` function parses these strings and implements the logic required
          to transform the dictionary of options (stored as strings) into shapes
          with associated properties.

    For creator user:
        * The creator user implements the `make` function (see above)
        * The class define the internal representation of a components
        * The class provides the interfaces for the component (creator user)

    Default Options:
        * pos_x/_y: '0.0um' -- The x/y position of the center of the QComponent.
        * orientation: '0.0' -- The primary direction in degrees of the QComponent.
          Expressed counter-clockwise orientation.
        * chip: 'main' -- Chip holding the QComponent.
        * layer: '1' -- Manufacturing layer used for the QComponent.

        Nested default options can be overwritten with the update function.
        The following code demonstrates how the update works.

        .. code-block:: python
            :linenos:

            from qiskit_metal import Dict
            default = Dict(
                a=1,
                b=2,
                c=Dict(
                    d=3,
                    e=4,
                    f=Dict(
                        g=6,
                        h=7
                    )
                )
            )
            overwrite = Dict(
                a=10,
                b=20,
                c=Dict(
                    d=30,
                    f=Dict(
                        h=70
                    )
                ),
                z=33
            )
            default.update(overwrite)
            default

        >> {'a': 10, 'b': 20, 'c': {'d': 30, 'e': 4, 'f': {'g': 6, 'h': 70}}, 'z': 33}
    """
    # pylint: disable=too-many-instance-attributes

    default_options = Dict(pos_x='0.0um',
                           pos_y='0.0um',
                           orientation='0.0',
                           chip='main',
                           layer='1')
    """Default drawing options"""

    component_metadata = Dict()
    """Component metadata"""

    TOOLTIP = """QComponent"""

    options = {}
    """A dictionary of the component-designer-defined options.
    These options are used in the make function to create the QGeometry and QPins.
    All options should have string keys and preferrable string values.
    """

    # Dummy private attribute used to check if an instantiated object is
    # indeed a QComponent class. The problem is that the `isinstance`
    # built-in method fails when this module is reloaded.
    # Used by `is_component` to check.
    __i_am_component__ = True

    def __init__(self,
                 design: 'QDesign',
                 name: str = None,
                 options: Dict = None,
                 make=True,
                 component_template: Dict = None) -> Union[None, str]:
        """Create a new Metal component and adds it's default_options to the
        design.

        Args:
            design (QDesign): The parent design.
            name (str): Name of the component. Auto-named if possible.
            options (dict): User options that will override the defaults.  Defaults to None.
            make (bool): True if the make function should be called at the end of the init.
                Options be used in the make function to create the geometry.  Defaults to True.
            component_template (dict): User can overwrite the template options for the component
                that will be stored in the design, in design.template,
                and used every time a new component is instantiated.  Defaults to None.

        Raises:
            ValueError: User supplied design isn't a QDesign

        Note:  Information copied from QDesign class.
            self._design.overwrite_enabled (bool):
            When True - If the string name, used for component, already
            exists in the design, the existing component will be
            deleted from design, and new component will be generated
            with the same name and newly generated component_id,
            and then added to design.

            When False - If the string name, used for component, already
            exists in the design, the existing component will be
            kept in the design, and current component will not be generated,
            nor will be added to the design. The variable design.self.status
            will still be NotBuilt, as opposed to Initialization Successful.

            Either True or False - If string name, used for component, is NOT
            being used in the design, a component will be generated and
            added to design using the name.
        """

        # Make the id be None, which means it hasn't been added to design yet.
        self._id = None
        self._made = False

        self._component_template = component_template

        # Status: used to handle building of a component and checking if it succeeded or failed.
        self.status = 'Not Built'
        if not is_design(design):
            raise ValueError(
                "Error you did not pass in a valid Metal QDesign object as a '\
                'parent of this QComponent.")

        self._design = design  # reference to parent
        # pylint: disable=literal-comparison
        if self._delete_evaluation(name) == 'NameInUse':
            raise ValueError(
                f"{name} already exists! Please choose a different name for your new QComponent"
            )

        self._name = name
        self._class_name = self._get_unique_class_name()  # Full class name
        #: A dictionary of the component-designer-defined options.
        #: These options are used in the make function to create the QGeometry and QPins.
        #: All options should have string keys and preferrable string values.
        self.options = self.get_template_options(
            design=design, component_template=component_template)
        if options:
            self.options.update(options)

        # Parser for options
        self.p = ParsedDynamicAttributes_Component(self)
        # Should put this earlier so could pass in other error messages?
        self._error_message = ''
        if self._check_pin_inputs():
            self.logger.warning(self._error_message)
            return
        # Build and component internals

        #: Dictionary of pins. Populated by component designer in make function using `add_pin`.
        self.pins = Dict()

        #: Metadata allows a designer to store extra information or analysis results.
        self.metadata = Dict()

        # Add the component to the parent design
        self._id = self.design._get_new_qcomponent_id()  # Create the unique id
        self._add_to_design()  # Do this after the pin checking?

        #: Stores the latest status of the component. Values include:
        #: ``Initialization Successful``, ``Build Failed``, etc.
        self.status = 'Initialization Successful'

        # Used for short name, and renderers adding information to tables.
        self.a_metadata = self._gather_all_children_metadata()

        # Auto naming - add id to component based on type
        if name is None:
            prefix = self.a_metadata
            # limit names to 24 characters
            name_trunc = 24
            # if no prefix, use class name
            if "short_name" not in prefix:
                short_name = self.__class__.__name__[:name_trunc]
            else:
                short_name = prefix['short_name'][:name_trunc]
            name_id = self.design._get_new_qcomponent_name_id(short_name)
            # rename loop to make sure that no components manually named by the user conflicts
            while self.design.rename_component(self._id, short_name + "_" +
                                               str(name_id)) != 1:
                name_id = self.design._get_new_qcomponent_name_id(short_name)

        # Add keys for each type of table.  add_qgeometry() will update bool if the table is used.
        self.qgeometry_table_usage = Dict()
        self.populate_to_track_table_usage()

        # Make the component geometry
        if make:
            self.rebuild()

    @classmethod
    def _gather_all_children_options(cls) -> dict:
        """From the QComponent core class, traverse the child classes to
        gather the `default_options` for each child class.

        Collects the options starting with the basecomponent,
        and stepping through the children.
        Each child adds it's options to the base options.  If the
        key is the same, the option of the youngest child is used.

        Note: if keys are the same for child and grandchild,
        grandchild will overwrite child

        Init method.

        Returns:
            dict: options from all children
        """

        options_from_children = {}
        parents = inspect.getmro(cls)

        # len-2: generic "object" does not have default_options.
        for child in parents[len(parents) - 2::-1]:
            # The template default options are in a class dict attribute `default_options`.
            if hasattr(child, 'default_options'):
                options_from_children = {
                    **options_from_children,
                    **child.default_options
                }

        if qlibrary.core.qroute.QRoute in parents:
            options_from_children.pop("pos_x", None)
            options_from_children.pop("pos_y", None)
            options_from_children.pop("orientation", None)

        return options_from_children

    @classmethod
    def _gather_all_children_metadata(cls) -> dict:
        """From the QComponent core class, traverse the child classes to
        gather the component_metadata for each child class.

        Note: if keys are the same for child and grandchild, grandchild will overwrite child
        Init method.

        Returns:
            dict: Metadata from all children.
        """

        metadata_from_children = {}
        parents = inspect.getmro(cls)
        # Base.py is not expected to have component_metadata dict to add to design class.
        for child in parents[len(parents) - 2::-1]:
            # There is a developer agreement so the defaults will be in dict named component_metadata.
            if hasattr(child, 'component_metadata'):
                metadata_from_children = {
                    **metadata_from_children,
                    **child.component_metadata
                }

        return metadata_from_children

    @classmethod
    def _get_unique_class_name(cls) -> str:
        """Returns unique class name based on the module.

        Returns:
            str: Example: 'qiskit_metal.qlibrary.qubits.transmon_pocket.TransmonPocket'
        """
        return f'{cls.__module__}.{cls.__name__}'

    @classmethod
    def _register_class_with_design(cls, design: 'QDesign', template_key: str,
                                    component_template: Dict):
        """Init function to register a component class with the design when
        first instantiated. Registers the design template options.

        Args:
            design (QDesign): The parent design
            template_key (str): Key to use
            component_template (dict): Template of components to copy, with renderer options
        """
        # do not overwrite
        if template_key not in design.template_options:
            # if not component_template:
            #     component_template = cls._gather_all_children_options()
            children_options = cls._gather_all_children_options()
            options_template_renderer = {
                **children_options,
                **component_template
            }
            # design.template_options[template_key] = deepcopy(
            #     component_template)
            design.template_options[template_key] = deepcopy(
                options_template_renderer)

    @property
    def name(self) -> str:
        """Name of the component."""
        return self._name

    @name.setter
    def name(self, new_name: str):
        """Rename the component. Change the design dictionaries as well. handle
        components. Delete and remake.

        Returns:
            bool: True is successful, otherwise failure code
        """
        return_code = self.design.rename_component(self.id, new_name)
        if return_code is not True:
            logger.warning(
                f'In design_base.name, the new_name={new_name} was not set. ')
        return return_code

    @property
    def design(self) -> 'QDesign':
        """Return a reference to the parent design object.

        Returns:
            QDesign: design
        """
        return self._design

    @property
    def class_name(self) -> str:
        """Return the full name of the class: the full module name with the
        class name. e.g., qiskit_metal.qlibrary.qubits.TransmonPocket.

        Returns:
            str: Class name
        """
        return self._class_name

    @property
    def logger(self) -> logging.Logger:
        """The Qiskit Metal Logger.

        Returns:
            logging.Logger: Logger
        """
        return self._design.logger

    @property
    def pin_names(self) -> set:
        """The names of the pins.

        Returns:
            set: Set of pin names
        """
        return set(self.pins.keys())

    @property
    # pylint: disable=invalid-name
    def id(self) -> int:
        """The unique id of component within a design.

        Returns:
            int: Component id
        """
        return self._id

    def _add_to_design(self):
        """Add self to design objects dictionary.

        Method will obtain an unique id for the component within a
        design, THEN add itself to design.
        """
        # pylint: disable=protected-access
        self.design._components[self.id] = self
        self.design.name_to_id[self.name] = self._id

    @classmethod
    def get_template_options(cls,
                             design: 'QDesign',
                             component_template: Dict = None,
                             logger_: logging.Logger = None,
                             template_key: str = None) -> Dict:
        """Creates template options for the Metal Component class required for
        the class to function, based on the design template; i.e., be created,
        made, and rendered. Provides the blank option structure required.

        The options can be extended by plugins, such as renderers.

        Args:
            design (QDesign): Design class. Should be the class, not the instance.
            component_template (Dict): Template options to overwrite the class ones (default: None)
            logger_ (logging.Logger): A logger for errors.  Defaults to None.
            template_key (str): The template key identifier. If None, then uses
                cls._get_unique_class_name().  Defaults to None.

        Returns:
            Dict: dictionary of default options based on design template.
        """
        # get key for templates
        if template_key is None:
            template_key = cls._get_unique_class_name()

        renderer_key_values = cls._get_table_values_from_renderers(design)
        # Think
        if component_template is not None:
            renderer_and_component_template = {
                **renderer_key_values,
                **component_template
            }
        else:
            renderer_and_component_template = renderer_key_values

        if template_key not in design.template_options:
            cls._register_class_with_design(design, template_key,
                                            renderer_and_component_template)

        if template_key not in design.template_options:
            logger_ = logger_ or design.logger
            if logger_:
                logger_.error(
                    f'ERROR in the creating component {cls.__name__}!\nThe default '
                    f'options for the component class {cls.__name__} are missing'
                )

        # Specific object template options
        template_options = deepcopy(Dict(design.template_options[template_key]))

        return template_options

    def _delete_evaluation(self, check_name: str = None):
        """When design.overwrite_enabled, the user is allowed to delete an
        existing component within the design if the name is being used.

        Args:
            check_name (str, optional): Name of new component.  Defaults to None.

        Returns:
            string: Return 'NameInUse' if overwrite flag is False and
            check_name is already being used within design.
            Otherwise return None.
        """
        answer = self._is_name_used(check_name)
        if self._design.overwrite_enabled and answer:
            self._design.delete_component(check_name)
        elif answer:
            logger.warning(
                f'The QComponent name `{check_name}` is already in use, '
                f'by a component (with QComponent id={answer}).\n'
                f'QComponent NOT made, nor added to the design. \n'
                'To force overwrite a QComponent with an existing name '
                'use the flag:\n`design.overwrite_enabled = True`.')
            return 'NameInUse'
        return None

    def make(self):
        """The make function implements the logic that creates the geometry
        (poly, path, etc.) from the qcomponent.options dictionary of
        parameters, and the adds them to the design, using
        qcomponent.add_qgeometry(...), adding in extra needed information, such
        as layer, subtract, etc.

        Use the qiskit_metal.draw module to create the geometry.

        **Note:**
            * This method should be overwritten by the children make function.
            * This function only contains the logic, the actual call to make the element is in
              rebuild()

        Raises:
            NotImplementedError: Overwrite this function by subclassing.
        """
        raise NotImplementedError()

    def to_script(self,
                  thin: bool = False,
                  is_part_of_chip: bool = False) -> Tuple:
        """

        Args:
            thin: If true then any key in the QComponent's options whose value
              is the same value as the default will not be included in the body
            is_part_of_chip: If true, body will not include header code


        Returns: Code that if copy-pasted into a .py file would generate
        an instance of this class with the same properties as the instance calling
        this function

        """

        def is_default_options(k):
            """ Returns true if option's key value is the same as the default value """
            temp_option = self.get_template_options(self.design)
            def_options = self.default_options

            if (k in def_options and def_options[k] == self.options[k]):
                return True

            if (k in temp_option and temp_option[k] == self.options[k]):
                return True

            return False

        module = self._get_unique_class_name()
        cls = '.'.join(module.split('.')[:-1])
        obj_name = module.split('.')[-1]

        ### constructing imports ###

        #header
        if not is_part_of_chip:
            header = """
from qiskit_metal import designs, MetalGUI

design = designs.DesignPlanar()

gui = MetalGUI(design)
            """
        else:
            header = ""

        # component import
        comp_import = f"""from {cls} import {obj_name}"""

        full_import = header + comp_import

        ### constructing qcomponent instantiation ###

        ## setting up options
        if thin:
            body_options = {}
            for k in self.options:
                if not is_default_options(k):
                    body_options[k] = self.options[k]
        else:
            body_options = self.options

        if len(body_options) < 1:
            str_options = ""
        else:
            pp = pprint.PrettyPrinter(width=41, compact=False)
            str_options = f"""options={pp.pformat(body_options)}"""

        ## setting up component-specific args
        # get init from child?
        to_ignore = {
            'self', 'name', 'design', 'make', 'kwargs', 'options', 'args'
        }
        class_signature = signature(self.__class__.__init__)

        failed = set()
        params = dict()
        str_params = ""
        for _, param in class_signature.parameters.items():
            if not param.name in to_ignore:
                param_name = param.name
                if param_name in self.__dict__:
                    param_name = param.name
                    param_val = self.__dict__[param.name]
                    if type(param_val) is str:
                        param_val = f"'{param_val}'"
                    params[param_name] = param_val
                elif '_' + param_name in self.__dict__:
                    priv_param_name = '_' + param_name
                    param_val = self.__dict__[priv_param_name]
                    if type(param_val) is str:
                        param_val = f"'{param_val}'"
                    params[param_name] = param_val
                else:
                    failed.add(param_name)
        for k, v in params.items():
            str_params += f"""
{k}={v},"""

        str_failed = ""
        if len(failed) > 0:
            str_failed += """

            # WARNING"""
        for k in failed:
            str_failed += f"""
#{k} failed to have a value"""

        ## setting up metadata
        if len(self.metadata) > 1:
            str_meta_d = f"""
{self.name}.meta = {self.metadata}
        """
        else:
            str_meta_d = ""

        ## cleaning up
        strname = self.name
        if not strname.isidentifier():
            if "-" in strname:
                strname = strname.replace("-", "")
            if not strname.isidentifier():
                strname = cls + str(random.randint(1000))

        other_args = ""
        if str_options != "":
            other_args += """,
""" + str_options

        if str_params != "":
            other_args += """,
""" + str_params

        ## setting up instantiation
        body = f"""
{str_failed}
{strname} = {obj_name}(
design,
name='{strname}'{other_args}
)
{str_meta_d}
"""

        return full_import, body

    def rebuild(self):
        """Builds the QComponent.

        This is the main action function of a
        QComponent, call it qc. It converts the qc.options into QGeometry with
        all of the required options, such as the geometry points, layer number,
        materials, etc. needed to render.

        The build clears the existing QGeometry and QPins and then calls the qc.make function,
        which is written by the component developer to implement the logic (using the metal.
        draw module) to convert the qc.options into the QGeometry.

        *Build status:*
        The function also sets the build status of the component.
        It sets to `failed` when the component is created, and then it sets to `good` when it is
        done with no errors. The user can also set other statuses, which can appear if the code
        fails to reach the final line of the build, where the build status is set to `good`.

        Raises:
            Exception: Component build failure
        """
        self.status = 'failed'
        try:
            if self._made:  # already made, just remaking
                self.design.qgeometry.delete_component_id(self.id)

                # pylint: disable=protected-access
                self.design._delete_all_pins_for_component(self.id)

            self.make()
            self._made = True
            self.status = 'good'

            self.design.build_logs.add_success(
                f"{str(datetime.now())} -- Component: {self.name} successfully built"
            )

        except Exception as error:
            self.logger.error(
                f'ERROR in building component name={self.name}, error={error}')
            self.design.build_logs.add_error(
                f"{str(datetime.now())} -- Component: {self.name} failed with error\n: {error}"
            )
            raise error

    def delete(self):
        """Delete the QComponent.

        Removes QGeometry, QPins, etc. from the design.
        """
        self.design.delete_component(self.name)

    # Maybe still should be fine as any values will be in component options still?
    # Though the data table approach and rendering directly via shapely could lead to problem
    # with variable use
    def parse_value(
        self, value: Union[Any, List, Dict, Iterable]
    ) -> Union[Any, List, Dict, Iterable]:
        """Parse a string, mappable (dict, Dict), iterable (list, tuple) to
        account for units conversion, some basic arithmetic, and design
        variables. This is the main parsing function of Qiskit Metal.

        Args:
            value (str): String to parse *or*
            variable_dict (dict): dict pointer of variables

        Return:
            str, float, list, tuple, or ast eval: Parse value

        Handled Inputs:
            Strings:
                Strings of numbers, numbers with units; e.g., '1', '1nm', '1 um'
                    Converts to int or float.
                    Some basic arithmetic is possible, see below.

                Strings of variables 'variable1'.
                    Variable interpretation will use string method
                    isidentifier 'variable1'.isidentifier()`

            Dictionaries:
                Returns ordered `Dict` with same key-value mappings, where the values have
                been subjected to parse_value.

            Itterables(list, tuple, ...):
                Returns same kind and calls itself `parse_value` on each element.

            Numbers:
                Returns the number as is. Int to int, etc.


        Arithmetic:
            Some basic arithmetic can be handled as well, such as `'-2 * 1e5 nm'`
            will yield float(-0.2) when the default units are set to `mm`.

        Default units:
            User units can be set in the design. The design will set config.DEFAULT.units

        Examples:
            See the docstring for this module.
                >> ?qiskit_metal.toolbox_metal.parsing
        """
        return self.design.parse_value(value)

    def parse_options(self, options: Dict = None) -> Dict:
        """Parse the options, converting string into interpreted values. Parses
        units, variables, strings, lists, and dictionaries. Explained by
        example below.

        Args:
            options (dict) : If left None, then self.options is used.  Defaults to None.

        Returns:
            dict: Parsed value

        Calls `self.design.parse_options`.

        See `self.parse_value` for more information.
        """
        return self.design.parse_value(options if options else self.options)

    def _is_name_used(self, check_name: str) -> int:
        """Used to check if name of component already exists.

        Args:
            check_name (str): Name which user requested to apply to current component.

        Returns:
            int: 0 if does not exist, otherwise
            component-id of component which is already using the name.

        Warning: If user has used this text version of the component name already,
        warning will be given to user.
        """

        if check_name in self._design.name_to_id:
            component_id = self._design.name_to_id[check_name]
            # if not self._overwrite_flag:
            #    logger.warning(f"Called _is_name_used, "
            #                   f"component_id({check_name}, id={component_id})"
            #                   " is already being used in design.")
            return component_id
        return 0

####################################################################################
# Functions for handling of pins

    def add_pin(
            self,
            name: str,  # Should be static based on component designer's choice
            points: np.ndarray,
            width: float,
            input_as_norm: bool = False,
            chip: str = None,
            gap: float = None):  # gap defaults to 0.6 * width
        """Adds a pin from two points which are normal/tangent to the intended
        plane of the pin. The normal should 'point' in the direction of
        intended connection. Adds the new pin as a subdictionary to parent
        component's pins dictionary.

        Args:
            * name (str): name of the pin
            * points (numpy.ndarray): [[x1,y1],[x2,y2]] for the normal/tangent line
            * width (float): the width of the intended connection (eg. qubit bus pad arm)
            * input_as_norm (bool): Indicates if the points are tangent or normal to the pin plane.
              Defaults to False.. Make True for normal.
            * parent (Union[int,]): The id of the parent component.
            * chip (str): the name of the chip the pin is located on. Defaults to None, which is 
            converted to self.options.chip.
            * gap (float): the dielectric gap of the pin for the purpose of representing as a port
              for simulations.  Defaults to None which is converted to 0.6 * width.

        Dictionary containing pins information:
            * points (numpy.ndarray) - two (x,y) points which represent the edge of the pin for
              another component to attach to (eg. the edge of a CPW TL)
            * middle (numpy.ndarray) - an (x,y) which represents the middle of the points above,
              where the pin is represented.
            * normal (numpy.ndarray) - the normal vector of the pin, pointing in direction of
              intended connection
            * tangent (numpy.ndarray) - 90 degree rotation of normal vector
            * width (float) - the width of the pin
            * chip (str) - the chip the pin is on
            * parent_name - the id of the parent component
            * net_id - net_id of the pin if connected to another pin.  Defaults to 0, indicates
              not connected))

        ::

            * = pin
            . = outline of component

        ---> = the list being passed in as 'points' [[x1,y1],[x2,y2]]


        normal vector

        ::

            ..........
                     .
            --------->*
                     .
            ..........

        tangent vector

        ::

            ..........^
                     .|
                     .*
                     .|
            ..........|
        """
        assert len(points) == 2

        if gap is None:
            gap = width * 0.6
        if chip is None:
            chip = self.options.chip

        rounding_val = self.design.template_options['PRECISION']
        points = np.around(
            points, rounding_val)  #Need points to remain as shapely geom?

        if input_as_norm:
            middle_point = points[1]
            vec_normal = points[1] - points[0]
            vec_normal /= np.linalg.norm(vec_normal)

            s_point = np.round(Vector.rotate(
                vec_normal, (np.pi / 2))) * width / 2 + points[1]
            e_point = np.round(Vector.rotate(
                vec_normal, -(np.pi / 2))) * width / 2 + points[1]
            points = [s_point, e_point]
            tangent_vector = Vector.rotate(vec_normal, np.pi / 2)

        else:
            vec_dist, tangent_vector, vec_normal = draw.Vector.two_points_described(
                points)
            middle_point = np.sum(points, axis=0) / 2.
            width = np.linalg.norm(vec_dist)

        pin_dict = Dict(
            points=points,
            middle=np.around(middle_point, rounding_val),
            normal=np.around(vec_normal, rounding_val),
            tangent=np.around(tangent_vector, rounding_val),
            width=np.around(width, rounding_val),
            gap=np.around(gap, rounding_val),
            chip=chip,
            parent_name=self.id,
            net_id=0,
            # Place holder value for potential future property (auto-routing cpw with
            # length limit)
            length=0)

        self.pins[name] = pin_dict

    def get_pin(self, name: str) -> Dict:
        """Interface for components to get pin data.

        Args:
            name (str): Name of the desired pin.

        Returns:
            dict: Returns the data of the pin, see make_pin() for what those values are.
        """

        return self.pins[name]

    def _check_pin_inputs(self):
        """Checks that the pin_inputs are valid, sets an error message
        indicating what the error is if the inputs are not valid. Checks
        regardless of user passing the component name or component id (probably
        a smoother way to do this check).

        3 Error cases:
            - Component does not exist
            - Pin does not exist
            - Pin is already attached to something

        Returns:
            str: Status test, or None
        """
        # Add check for if user inputs nonsense?
        # pylint: disable=protected-access
        false_component = False
        false_pin = False
        pin_in_use = False
        for pin_check in self.options.pin_inputs.values():
            component = pin_check['component']
            pin = pin_check['pin']
            if isinstance(component, str):
                if component not in self.design.components:
                    false_component = True
                elif pin not in self.design.components[component].pins:
                    false_pin = True
                elif self.design.components[component].pins[pin].net_id:
                    pin_in_use = True
            elif isinstance(component, int):
                if component not in self.design._components:
                    false_component = True
                elif pin not in self.design._components[component].pins:
                    false_pin = True
                elif self.design._components[component].pins[pin].net_id:
                    pin_in_use = True
            # Should modify to allow for multiple error messages to be returned.
            if false_component:
                self._error_message = (
                    f'Component {component} does not exist. {self.name} has not been built. '
                    'Please check your pin_input values.')
                return 'Component Does Not Exist'
            if false_pin:
                self._error_message = (
                    f'Pin {pin} does not exist in component {component}. '
                    f'{self.name} has not been built. Please check your pin_input values.'
                )
                return 'Pin Does Not Exist'
            if pin_in_use:
                self._error_message = (
                    f'Pin {pin} of component {component} is already in use. '
                    f'{self.name} has not been built. Please check your pin_input values.'
                )
                return 'Pin In Use'
        return None

    # This method does not appear to be being used anywhere.
    def connect_components_already_in_design(self, pin_name_self: str,
                                             comp2_id: int,
                                             pin2_name: str) -> int:
        """WARNING: Do NOT use this method during generation of component instance.
            This method is expecting self to be added to design._components dict.  More importantly,
            the unique id of self component needs to be in design._components dict.

        Args:
            pin_name_self (str): Name of pin within the component.
            comp2_id (int): Component within design, but not self.
            pin2_name (str): The pin of comp2_id that pin_name_self will connect to.

        Returns:
            int: A unique net_id for the connection.
        """
        # pylint: disable=protected-access

        net_id_rtn = 0

        if self.id not in self.design._components:
            # Component not in design.
            logger.warning(
                f'No connection made. Component_id {self.id} not in design.')
            return net_id_rtn

        if comp2_id not in self.design._components:
            # Component not in design.
            logger.warning(
                f'No connection made. Component_id {comp2_id} not in design.')
            return net_id_rtn

        if self.design._components[self._id].pins[pin_name_self].net_id:
            # Pin already in use.
            logger.warning(
                f'Component_id {self._id} not connected.  The pin '
                f'{pin_name_self} is already connected to something else.')
            return net_id_rtn

        if self.design._components[comp2_id].pins[pin2_name].net_id:
            # Pin already in use.
            logger.warning(
                f'Component_id {comp2_id} not connected.  The pin '
                f'{pin2_name} is already connected to something else.')
            return net_id_rtn

        net_id_rtn = self.design.connect_pins(self.id, pin_name_self, comp2_id,
                                              pin2_name)

        return net_id_rtn

########################################################################

    def add_dependency(self, parent: str, child: str):
        """Add a dependency between one component and another. Calls parent
        design.

        Args:
            parent (str): The component on which the child depends
            child (str): The child cannot live without the parent.
        """
        self.design.add_dependency(parent, child)

##########################################
# QGeometry

    def add_qgeometry(
            self,
            kind: str,
            geometry: dict,
            subtract: bool = False,
            helper: bool = False,
            layer: Union[int, str] = None,  # chip will be here
            chip: str = None,
            **kwargs):
        r"""Add QGeometry.

        Takes any additional options in options.

        Args:
            kind (str): The kind of QGeometry, such as 'path', 'poly', etc.
                        All geometry in the dictionary should have the same kind,
                        such as Polygon or LineString.
            geometry (Dict[BaseGeometry]): Key-value pairs of name of the geometry
                you want to add and the value should be a shapely geometry object, such
                as a Polygon or a LineString.
            subtract (bool): Subtract from the layer.  Defaults to False.
            helper (bool): Is this a helper object. If true, subtract must be false
                           Defaults to False.
            layer (int, str): The layer to which the set of QGeometry will belong
                              Defaults to None, which is converted to self.options.chip.
            chip (str): Chip name. Defaults to None, which is converted to 
            self.options.chip.
            kwargs (dict): Parameters dictionary

        Assumptions:
            * Assumes all geometry in the `geometry` argument are homogeneous in kind;
              i.e., all lines or polys etc.
        """
        # assert (subtract and helper) == False, "The object can't be a subtracted helper. Please"\
        #    " choose it to either be a helper or a a subtracted layer, but not both. Thank you."

        if layer is None:
            layer = self.options.layer
        if chip is None:
            chip = self.options.chip

        if kind in self.qgeometry_table_usage.keys():
            self.qgeometry_table_usage[kind] = True
        else:
            self.logger.warning(
                f'Component with classname={self.class_name} does not know about '
                f'table name "{kind}".')

        renderer_key_values = self._get_specific_table_values_from_renderers(
            kind)
        for key in renderer_key_values:
            if key in self.options:
                renderer_key_values[key] = deepcopy(self.options[key])

        # # if not already in kwargs, add renderer information to it.
        renderer_and_options = {**renderer_key_values, **kwargs}

        # When self.options is instantiated, the template_options are populated.
        # renderer_and_options = {**self.options, **kwargs}

        self.design.qgeometry.add_qgeometry(kind,
                                            self.id,
                                            geometry,
                                            subtract=subtract,
                                            helper=helper,
                                            layer=layer,
                                            chip=chip,
                                            **renderer_and_options)

    def _get_specific_table_values_from_renderers(self, kind: str) -> Dict:
        """Populate a dict to combine with options for the qcomponent.

        Based on kind, which the table name, the component-developer denotes in the metadata,
        assume those qgeometry.tables are used for the component. The method
        will search a dict populated by all the renderers during their init.

        Args:
            kind (str): Name of table, like junction, path, or poly.

        Returns:
            Dict: key is column names for tables, value is data for the column.
        """
        all_renderers_key_value = dict()

        # design.renderer_defaults_by_table[table_name][renderer_name][column_name]
        if kind in self.design.renderer_defaults_by_table:
            for name_renderer, renderer_data in self.design.renderer_defaults_by_table[
                    kind].items():
                if len(renderer_data) > 0:
                    for col_name, col_value in renderer_data.items():
                        render_col_name = f'{name_renderer}_{col_name}'
                        all_renderers_key_value[render_col_name] = col_value

        return all_renderers_key_value

    @classmethod
    def _get_table_values_from_renderers(cls, design: 'QDesign') -> Dict:
        """Populate a dict to combine with options for the qcomponent.

        Based on tables the component-developer denotes in the metadata,
        assume those qgeometry.tables are used for the component. The method
        will search a dict populated by all the renderers during their init.

        Returns:
            Dict: key is column names for tables, value is data for the column.
        """
        metadata_dict = cls._gather_all_children_metadata()
        tables_list = design.get_list_of_tables_in_metadata(metadata_dict)

        all_renderers_key_value = dict()
        # design.renderer_defaults_by_table[table_name][renderer_name][column_name]
        for table in tables_list:
            if table in design.renderer_defaults_by_table:
                for name_renderer, renderer_data in design.renderer_defaults_by_table[
                        table].items():
                    if len(renderer_data) > 0:
                        for col_name, col_value in renderer_data.items():
                            render_col_name = f'{name_renderer}_{col_name}'
                            all_renderers_key_value[render_col_name] = col_value
        return all_renderers_key_value

######################################

    def __repr__(self, *args):
        # pylint: disable=invalid-name

        b = '\033[95m\033[1m'
        b1 = '\033[94m\033[1m'
        e = '\033[0m'

        # id = {hex(id(self))}
        # options = pprint.pformat(self.options)

        options = format_dict_ala_z(self.options)
        text = f"{b}name:    {b1}{self.name}{e}\n"\
            f"{b}class:   {b1}{self.__class__.__name__:<22s}{e}\n"\
            f"{b}options: {e}\n{options}\n"\
            f"{b}module:  {b1}{self.__class__.__module__}{e}\n"\
            f"{b}id:      {b1}{self.id}{e}\n"
        return text


############################################################################
# Geometry handling of created qgeometry

    @property
    def qgeometry_types(self) -> List[str]:
        """Get a list of the names of the element tables.

        Returns:
            List[str]: Name of element table or type; e.g., 'poly' and 'path'
        """
        return self.design.qgeometry.get_element_types()

    def qgeometry_dict(  # pylint: disable=inconsistent-return-statements
            self, element_type: str) -> Dict_[str, BaseGeometry]:
        """Returns a dict of element qgeometry (shapely geometry) of the
        component as a python dict, where the dict keys are the names of the
        qgeometry and the corresponding values are the shapely geometries.

        Args:
            element_type (str): Name of element table or type; e.g., 'poly' and 'path'

        Returns:
            List[BaseGeometry]: Geometry diction or None if an error in the name of the element
            type (ie. table)
        """

        if element_type == 'all' or self.design.qgeometry.check_element_type(
                element_type):
            return self.design.qgeometry.get_component_geometry_dict(
                self.name, element_type)

    def qgeometry_list(  # pylint: disable=inconsistent-return-statements
            self,
            element_type: str = 'all') -> List[BaseGeometry]:
        """Returns a list of element qgeometry (shapely geometry) of the
        component as a python list of shapely geometries.

        Args:
            element_type (str): Name of element table or type; e.g., 'poly' and 'path'.
                                Can also specify all

        Returns:
            List[BaseGeometry]: Geometry list or None if an error in the name of the element type
            (ie. table)
        """

        if element_type == 'all' or self.design.qgeometry.check_element_type(
                element_type):
            return self.design.qgeometry.get_component_geometry_list(
                self.name, element_type)

    def qgeometry_table(  # pylint: disable=inconsistent-return-statements
            self, element_type: str) -> pd.DataFrame:
        """Returns the entire element table for the component.

        Args:
            element_type (str): Name of element table or type; e.g., 'poly' and 'path'

        Returns:
            pd.DataFrame: Element table for the component or None if an error in the name of
            the element type (ie. table)
        """

        if element_type == 'all' or self.design.qgeometry.check_element_type(
                element_type):
            return self.design.qgeometry.get_component(self.name, element_type)

    def qgeometry_bounds(self) -> Tuple:
        """Fetched the component bound dict_value.

        Returns:
            tuple: containing (minx, miny, maxx, maxy) bound values for the bounds of the
            component as a whole.

        Uses:
            design.qgeometry.get_component_bounds
        """
        bounds = self.design.qgeometry.get_component_bounds(self.name)
        return bounds

    def qgeometry_plot(self,
                       ax: 'matplotlib.axes.Axes' = None,
                       plot_kw: dict = None) -> List:
        """Draw all the qgeometry of the component (polys and path etc.)

        Args:
            ax (matplotlib.axes.Axes): Matplotlib axis to draw on.  Defaults to None.
                                       When None, it gets the current axis.
            plot_kw (dict): Parameters dictionary.

        Returns:
            List: The list of qgeometry draw

        Example use:
            Suppose you had a component called q1:

                fig, ax = draw.mpl.figure_spawn()
                q1.qgeometry_plot(ax)
        """
        qgeometry = self.qgeometry_list()
        plot_kw = {}
        draw.mpl.render(qgeometry, ax=ax, kw=plot_kw)
        return qgeometry

    def populate_to_track_table_usage(self) -> None:
        """Use the element_handler to get a list of all the table names used in
        QGeometry.

        The dict qgeometry_able_usage should get updated by
        add_qgeometry(). This dict is used to get a summary tables used
        for this component.
        """
        for table_name in self.design.qgeometry.tables.keys():
            self.qgeometry_table_usage[table_name] = False



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Concentric Transmon, as shown in Phys. Rev. Appl. 10, 034050 (2018).
"""
from math import sin, cos
import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import BaseQubit


class TransmonConcentricType2(BaseQubit):
    """The base `TrasmonConcentric` class
    Inherits `BaseQubit` class.
    Metal transmon object consisting of a circle surrounding by a concentric
    ring. There are two Josephson Junction connecting the circle to the ring;
    one at the south end and one at the north end. There is a readout resonator.

    .. TODO: add TransmonConcentricType2.png when available

    .. meta::
        :description: Transmon Concentric Type 2

    BaseQubit Default Options:
        * connection_pads: empty Dict -- the dictionary which contains all active
          connection lines for the qubit.
        * _default_connection_pads: empty Dict -- the default values for the
          (if any) connection lines of the qubit.
    Default Options:
        * width: '1000um' -- Width of transmon pocket
        * height: '1000um' -- Height of transmon pocket
        * rad_o: '170um' -- Outer radius
        * rad_i: '115um' -- Inner radius
        * gap: '35um' -- Radius of gap between two pads
        * jj_w: '10um' -- Josephson Junction width
        * res_s: '100um' -- Space between top electrode and readout resonator
        * res_ext: '100um' -- Extension of readout resonator in x-direction
          beyond midpoint of transmon
        * fbl_rad: '100um' -- Radius of the flux bias line loop
        * fbl_sp: '100um' -- Spacing between metal pad and flux bias loop
        * fbl_gap: '80um' -- Space between parallel lines of the flux bias loop
        * fbl_ext: '300um' -- Run length of flux bias line between circular
          loop and edge of pocket
        * pocket_w: '1500um' -- Transmon pocket width
        * pocket_h: '1000um' -- Transmon pocket height
        * cpw_width: '10.0um' -- Width of the readout resonator and flux bias line
        * layer: '1' -- default layer
    """

    # default drawing options
    default_options = Dict(pos_x='0mm',
                           pos_y='0mm',
                           orientation='0',
                           rad_outer='2mm',
                           rad_inner='1mm',
                           gap='0.5mm',
                           jj_w='0.2mm',
                           finger_NE_width='0.1mm',
                           finger_NE_length='0.5mm',
                           coupler_NE_width='0.1mm',
                           coupler_NE_gap='0.1mm',
                           finger_NW_width='0.1mm',
                           finger_NW_length='0.5mm',
                           coupler_NW_width='0.1mm',
                           coupler_NW_gap='0.1mm',
                           finger_SW_width='0.1mm',
                           finger_SW_length='0.5mm',
                           coupler_SW_width='0.1mm',
                           coupler_SW_gap='0.1mm',
                           finger_E_width='0.1mm',
                           finger_E_length='0.5mm',
                           box_E_width='0.7mm',
                           box_E_length='0.7mm',
                           finger_N_width='0.1mm',
                           finger_N_length='2mm',
                           triangle_base='2mm',
                           triangle_height='1mm',
                           pocket_w='10mm',
                           pocket_h='10mm',
                           layer='1')
    """Default drawing options"""

    def make(self):
        """Convert self.options into QGeometry."""
        p = self.parse_options()  # Parse the string options into numbers
        # draw the concentric pad regions
        outer_pad = draw.Point(0, 0).buffer(p.rad_outer)
        space = draw.Point(0, 0).buffer((p.gap + p.rad_inner))
        cutout = draw.rectangle(8.0 * p.finger_N_width, p.finger_N_length, 0.0,
                                p.rad_inner + 0.5 * p.finger_N_length)
        outer_pad = draw.subtract(outer_pad, space)
        outer_pad = draw.subtract(outer_pad, cutout)
        inner_pad = draw.Point(0, 0).buffer(p.rad_inner)
        # draw the Josephson Junction
        JJ = draw.LineString([
            (p.rad_inner * cos(45 * 3.14159 / 180),
             -1.0 * p.rad_inner * sin(45 * 3.14159 / 180)),
            ((p.rad_inner + p.gap) * cos(45 * 3.14159 / 180),
             -1.0 * (p.rad_inner + p.gap) * sin(45 * 3.14159 / 180))
        ])
        # draw the northeast finger
        finger_NE = draw.LineString([
            (p.rad_outer * cos(45 * 3.14159 / 180),
             p.rad_outer * sin(45 * 3.14159 / 180)),
            ((p.rad_outer + p.finger_NE_length) * cos(45 * 3.14159 / 180),
             (p.rad_outer + p.finger_NE_length) * sin(45 * 3.14159 / 180))
        ])
        # draw the coupling resonator to the northeast finger
        r = (p.rad_outer + 0.5 * p.finger_NE_length + 0.5 *
             (0.5 * p.finger_NE_length + p.coupler_NE_gap + p.coupler_NE_width))
        coupler_NE = draw.rectangle(
            2.0 * p.coupler_NE_width + p.finger_NE_width + 2 * p.coupler_NE_gap,
            0.5 * p.finger_NE_length + p.coupler_NE_gap + p.coupler_NE_width,
            r * cos(45 * 3.14159 / 180), r * sin(45 * 3.14159 / 180))
        coupler_NE = draw.rotate(coupler_NE,
                                 -45,
                                 origin=(r * cos(45 * 3.14159 / 180),
                                         r * sin(45 * 3.14159 / 180)))
        l = p.rad_outer + 0.5 * p.finger_NE_length + 0.5 * (
            0.5 * p.finger_NE_length + p.coupler_NE_gap)
        coupler_NE_cut = draw.rectangle(
            p.finger_NE_width + 2.0 * p.coupler_NE_gap,
            0.5 * p.finger_NE_length + p.coupler_NE_gap +
            0.05 * p.finger_NE_length, l * cos(45 * 3.145159 / 180),
            l * sin(45 * 3.14159 / 180))
        coupler_NE_cut = draw.rotate(coupler_NE_cut,
                                     -45,
                                     origin=(l * cos(45 * 3.14159 / 180),
                                             l * sin(45 * 3.14159 / 180)))
        coupler_NE = draw.subtract(coupler_NE, coupler_NE_cut)
        # draw the northwest finger
        finger_NW = draw.LineString([
            (-1.0 * p.rad_outer * cos(45 * 3.14159 / 180),
             p.rad_outer * sin(45 * 3.14159 / 180)),
            (-1.0 * (p.rad_outer + p.finger_NW_length) *
             cos(45 * 3.14159 / 180),
             (p.rad_outer + p.finger_NW_length) * sin(45 * 3.14159 / 180))
        ])
        # draw the coupling resonator to the northwest finger
        r_nw = (
            p.rad_outer + 0.5 * p.finger_NW_length + 0.5 *
            (0.5 * p.finger_NW_length + p.coupler_NW_gap + p.coupler_NW_width))
        coupler_NW = draw.rectangle(
            2.0 * p.coupler_NW_width + p.finger_NW_width + 2 * p.coupler_NW_gap,
            0.5 * p.finger_NW_length + p.coupler_NW_gap + p.coupler_NW_width,
            -1.0 * r_nw * cos(45 * 3.14159 / 180),
            r_nw * sin(45 * 3.14159 / 180))
        coupler_NW = draw.rotate(coupler_NW,
                                 45,
                                 origin=(-1.0 * r_nw * cos(45 * 3.14159 / 180),
                                         r_nw * sin(45 * 3.14159 / 180)))
        l_nw = p.rad_outer + 0.5 * p.finger_NW_length + 0.5 * (
            0.5 * p.finger_NW_length + p.coupler_NW_gap)
        coupler_NW_cut = draw.rectangle(
            p.finger_NW_width + 2.0 * p.coupler_NW_gap,
            0.5 * p.finger_NW_length + p.coupler_NW_gap +
            0.05 * p.finger_NW_length, -1.0 * l_nw * cos(45 * 3.145159 / 180),
            l_nw * sin(45 * 3.14159 / 180))
        coupler_NW_cut = draw.rotate(
            coupler_NW_cut,
            45,
            origin=(-1.0 * l_nw * cos(45 * 3.14159 / 180),
                    l_nw * sin(45 * 3.14159 / 180)))
        coupler_NW = draw.subtract(coupler_NW, coupler_NW_cut)
        # draw the southwest finger
        finger_SW = draw.LineString([
            (-1.0 * p.rad_outer * cos(45 * 3.14159 / 180),
             -1.0 * p.rad_outer * sin(45 * 3.14159 / 180)),
            (-1.0 * (p.rad_outer + p.finger_SW_length) *
             cos(45 * 3.14159 / 180), -1.0 *
             (p.rad_outer + p.finger_SW_length) * sin(45 * 3.14159 / 180))
        ])
        # draw the coupling resonator to the southwest finger
        r_sw = (
            p.rad_outer + 0.5 * p.finger_SW_length + 0.5 *
            (0.5 * p.finger_SW_length + p.coupler_SW_gap + p.coupler_SW_width))
        coupler_SW = draw.rectangle(
            2.0 * p.coupler_SW_width + p.finger_SW_width + 2 * p.coupler_SW_gap,
            0.5 * p.finger_SW_length + p.coupler_SW_gap + p.coupler_SW_width,
            -1.0 * r_sw * cos(45 * 3.14159 / 180),
            -1.0 * r_sw * sin(45 * 3.14159 / 180))
        coupler_SW = draw.rotate(coupler_SW,
                                 135,
                                 origin=(-1.0 * r_sw * cos(45 * 3.14159 / 180),
                                         -1.0 * r_sw * sin(45 * 3.14159 / 180)))
        l_sw = p.rad_outer + 0.5 * p.finger_SW_length + 0.5 * (
            0.5 * p.finger_SW_length + p.coupler_SW_gap)
        coupler_SW_cut = draw.rectangle(
            p.finger_SW_width + 2.0 * p.coupler_SW_gap,
            0.5 * p.finger_SW_length + p.coupler_SW_gap +
            0.05 * p.finger_SW_length, -1.0 * l_sw * cos(45 * 3.145159 / 180),
            -1.0 * l_sw * sin(45 * 3.14159 / 180))
        coupler_SW_cut = draw.rotate(
            coupler_SW_cut,
            135,
            origin=(-1.0 * l_sw * cos(45 * 3.14159 / 180),
                    -1.0 * l_sw * sin(45 * 3.14159 / 180)))
        coupler_SW = draw.subtract(coupler_SW, coupler_SW_cut)
        # draw the east finger with the rectangular coupling pad
        finger_E = draw.LineString([(-1.0 * p.rad_outer, 0.0),
                                    (-1.0 * (p.rad_outer + p.finger_E_length),
                                     0.0)])
        box = draw.rectangle(
            p.box_E_width, p.box_E_length,
            -1.0 * (p.rad_outer + p.finger_E_length + 0.5 * p.box_E_width), 0.0)
        # draw the north finger with the triangular coupling pad
        finger_N = draw.LineString([(0.0, p.rad_inner),
                                    (0.0, p.rad_inner + p.finger_N_length)])
        # draw the top triangular coupling pad
        padtop_i = draw.rectangle(
            p.triangle_base, p.triangle_height, 0.0,
            p.rad_inner + p.finger_N_length + 0.5 * p.triangle_height)
        cut1 = draw.rotate(padtop_i,
                           45,
                           origin=(0.5 * p.triangle_base, p.rad_inner +
                                   p.finger_N_length + p.triangle_height))
        padtop = draw.subtract(padtop_i, cut1)
        cut2 = draw.rotate(padtop_i,
                           315,
                           origin=(-0.5 * p.triangle_base, p.rad_inner +
                                   p.finger_N_length + p.triangle_height))
        padtop = draw.subtract(padtop, cut2)
        padtop = draw.translate(padtop, 0.0, -.05 * p.triangle_height)
        # draw the transmon pocket bounding box
        # pocket = draw.rectangle(p.pocket_w, p.pocket_h)
        # Translate and rotate all shapes
        objects = [
            outer_pad, inner_pad, JJ, finger_NE, finger_NW, finger_SW, finger_E,
            box, finger_N, padtop, coupler_NE, coupler_NW, coupler_SW
        ]
        objects = draw.rotate(objects, p.orientation, origin=(0, 0))
        objects = draw.translate(objects, xoff=p.pos_x, yoff=p.pos_y)
        [
            outer_pad, inner_pad, JJ, finger_NE, finger_NW, finger_SW, finger_E,
            box, finger_N, padtop, coupler_NE, coupler_NW, coupler_SW
        ] = objects
        ##############################################################
        # Use the geometry to create Metal QGeometry
        geom_outer = {'poly1': outer_pad}
        geom_inner = {'poly2': inner_pad}
        geom_jj = {'poly3': JJ}
        geom_finger_NE = {'finger_NE': finger_NE}
        geom_finger_NW = {'finger_NW': finger_NW}
        geom_finger_SW = {'finger_SW': finger_SW}
        geom_finger_E = {'finger_E': finger_E}
        geom_box_E = {'box_E': box}
        geom_finger_N = {'finger_N': finger_N}
        geom_pad_top = {'pad_top': padtop}
        geom_coupler_NE = {'coupler_NE': coupler_NE}
        geom_coupler_NW = {'coupler_NW': coupler_NW}
        geom_coupler_SW = {'coupler_SW': coupler_SW}
        self.add_qgeometry('poly', geom_outer, layer=p.layer, subtract=False)
        self.add_qgeometry('poly', geom_inner, layer=p.layer, subtract=False)
        self.add_qgeometry('junction',
                           geom_jj,
                           layer=p.layer,
                           subtract=False,
                           width=p.jj_w)
        self.add_qgeometry('path',
                           geom_finger_NE,
                           layer=p.layer,
                           subtract=False,
                           width=p.finger_NE_width)
        self.add_qgeometry('path',
                           geom_finger_NW,
                           layer=p.layer,
                           subtract=False,
                           width=p.finger_NW_width)
        self.add_qgeometry('path',
                           geom_finger_SW,
                           layer=p.layer,
                           subtract=False,
                           width=p.finger_SW_width)
        self.add_qgeometry('path',
                           geom_finger_E,
                           layer=p.layer,
                           subtract=False,
                           width=p.finger_E_width)
        self.add_qgeometry('poly', geom_box_E, layer=p.layer, subtract=False)
        self.add_qgeometry('path',
                           geom_finger_N,
                           layer=p.layer,
                           subtract=False,
                           width=p.finger_N_width)
        self.add_qgeometry('poly', geom_pad_top, layer=p.layer, subtract=False)
        self.add_qgeometry('poly',
                           geom_coupler_NE,
                           layer=p.layer,
                           subtract=False)
        self.add_qgeometry('poly',
                           geom_coupler_NW,
                           layer=p.layer,
                           subtract=False)
        self.add_qgeometry('poly',
                           geom_coupler_SW,
                           layer=p.layer,
                           subtract=False)

        ###########################################################################
        # Add Qpin connections
        # define a function that both rotates and translates the qpin coordinates
        def qpin_rotate_translate(x):
            y = list(x)
            z = [0.0, 0.0]
            z[0] = y[0] * cos(p.orientation * 3.14159 / 180) - y[1] * sin(
                p.orientation * 3.14159 / 180)
            z[1] = y[0] * sin(p.orientation * 3.14159 / 180) + y[1] * cos(
                p.orientation * 3.14159 / 180)
            z[0] = z[0] + p.pos_x
            z[1] = z[1] + p.pos_y
            x = (z[0], z[1])
            return x

        # Northeast coupler
        qp1a_r = p.rad_outer + p.finger_NE_length + p.coupler_NE_gap
        qp1a = (qp1a_r * cos(45 * 3.14159 / 180),
                qp1a_r * sin(45 * 3.14159 / 180))
        qp1b_r = p.rad_outer + p.finger_NE_length + p.coupler_NE_gap + p.coupler_NE_width
        qp1b = (qp1b_r * cos(45 * 3.14159 / 180),
                qp1b_r * sin(45 * 3.14159 / 180))
        qp1a = qpin_rotate_translate(qp1a)
        qp1b = qpin_rotate_translate(qp1b)
        self.add_pin('pin1',
                     points=np.array([qp1a, qp1b]),
                     width=0.01,
                     input_as_norm=True)
        # Northwest coupler
        qp2a_r = p.rad_outer + p.finger_NW_length + p.coupler_NW_gap
        qp2a = (-1.0 * qp2a_r * cos(45 * 3.14159 / 180),
                qp2a_r * sin(45 * 3.14159 / 180))
        qp2b_r = p.rad_outer + p.finger_NW_length + p.coupler_NW_gap + p.coupler_NW_width
        qp2b = (-1.0 * qp2b_r * cos(45 * 3.14159 / 180),
                qp2b_r * sin(45 * 3.14159 / 180))
        qp2a = qpin_rotate_translate(qp2a)
        qp2b = qpin_rotate_translate(qp2b)
        self.add_pin('pin2',
                     points=np.array([qp2a, qp2b]),
                     width=0.01,
                     input_as_norm=True)
        # Southwest coupler
        qp3a_r = p.rad_outer + p.finger_SW_length + p.coupler_SW_gap
        qp3a = (-1.0 * qp3a_r * cos(45 * 3.14159 / 180),
                -1.0 * qp3a_r * sin(45 * 3.14159 / 180))
        qp3b_r = p.rad_outer + p.finger_SW_length + p.coupler_SW_gap + p.coupler_SW_width
        qp3b = (-1.0 * qp3b_r * cos(45 * 3.14159 / 180),
                -1.0 * qp3b_r * sin(45 * 3.14159 / 180))
        qp3a = qpin_rotate_translate(qp3a)
        qp3b = qpin_rotate_translate(qp3b)
        self.add_pin('pin3',
                     points=np.array([qp3a, qp3b]),
                     width=0.01,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=relative-beyond-top-level
# pylint: disable-msg=import-error
"""Unit tests"""

from .. import config
if config.is_building_docs():
    from .assertions import AssertionsMixin
    from .test_analyses import TestAnalyses
    from .test_default_options import TestDefautOptions
    from .test_designs import TestDesign
    from .test_draw import TestDraw
    from .test_gui_basic import TestGUIBasic
    from .test_qgeometries import TestElements
    from .test_qlibrary_1_instantiate import TestComponentInstantiation
    from .test_qlibrary_2_options import TestComponentOptions
    from .test_qlibrary_3_functionality import TestComponentFunctionality
    from .test_renderers import TestRenderers
    from .test_speed import TestSpeed
    from .test_toolbox_metal import TestToolboxMetal
    from .test_toolbox_python import TestToolboxPython



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Simply utility functions to improve QOL of QM developers and QM users."""

import logging
import os
import re
import sys
import traceback
import warnings
from copy import deepcopy
from typing import Dict, List, TYPE_CHECKING, Tuple, Callable, Union
import inspect

import pandas as pd

from qiskit_metal.draw import Vector
from qiskit_metal.toolbox_metal.exceptions import InputError

if TYPE_CHECKING:
    from qiskit_metal import logger

__all__ = [
    'copy_update', 'dict_start_with', 'data_frame_empty_typed', 'clean_name',
    'enable_warning_traceback', 'get_traceback', 'print_traceback_easy',
    'log_error_easy', 'monkey_patch', 'can_write_to_path',
    'can_write_to_path_with_warning', 'toggle_numbers', 'bad_fillet_idxs',
    'compress_vertex_list', 'get_range_of_vertex_to_not_fillet'
]

####################################################################################
# Dictionary related


def copy_update(options, *args, deep_copy=True, **kwargs):
    """Utility funciton to merge two dictionaries.

    Args:
        options (object): Options
        deep_copy (bool): True to do a deep copy
        kwargs (dict): Dictionary of parameters

    Returns:
        dict: Merged dictionary
    """
    if deep_copy:
        options = deepcopy(options)
        options.update(*args, **kwargs)
    else:
        options = options.copy()
        options.update(*args, **kwargs)
    return options


def dict_start_with(my_dict, start_with, as_=list):
    """Case sensitive https://stackoverflow.com/questions/17106819/accessing-
    python-dict-values-with-the-key-start-characters.

    Args:
        my_dict (dict): The dictionary
        starts_with (str): String to check of
        as_ (type): A list of dict.  Defaults to list.

    Returns:
        list or dict: Parts of the dictionary with keys starting with the given text

    .. code-block:: python

        my_dict = {'name': 'Klauss', 'age': 26, 'Date of birth': '15th july'}
        dict_start_with(my_dict, 'Date')
    """
    if as_ == list:
        # start_with in k]
        return [v for k, v in my_dict.items() if k.startswith(start_with)]
    elif as_ == dict:
        return {k: v for k, v in my_dict.items() if k.startswith(start_with)}


# def display_options(*ops_names, options=None, find_dot_keys=True, do_display=True):
#     '''
#     Print html display of options dictionary by default `DEFAULT_OPTIONS`

#     Example use:
#     ---------------
#         display_options('make_transmon_pocket_v1', 'make_transmon_connector_v1')

#     or
#         dfs, html = display_options(Metal_Transmon_Pocket.__name__, do_display=False)
#     '''
#     # IDEA: display also ._hfss and ._gds etc. for those that have it and add to plugins
#     if options is None:
#         from .. import DEFAULT_OPTIONS
#         options = DEFAULT_OPTIONS

#     res = []
#     for keyname in ops_names:
#         if find_dot_keys:
#             names = list(filter(lambda x, match=keyname: x is match or
#                                 x.startswith(match+'.'), DEFAULT_OPTIONS.keys()))
#             names.sort()
#             for name in names:
#                 res += [pd.Series(options[name], name=name).to_frame()]
#         else:
#             res += [pd.Series(options[keyname], name=keyname).to_frame()]

#     from pyEPR.toolbox import display_dfs
#     res_html = display_dfs(*res, do_display=do_display)  #why not just directly call the function DataFrame_display_side_by_side(*args) ?
#     return res, res_html


def data_frame_empty_typed(column_types: dict):
    """Creates and empty DataFrame with dtypes for each column given by the
    dictionary.

    Args:
        column_types (dict): A key, dtype pairs

    Returns:
        DataFrame: An empty dataframe with the typed columns
    """
    df = pd.DataFrame()
    for name, dtype in column_types.items():
        df[name] = pd.Series(dtype=dtype)
    return df


def clean_name(text: str):
    """Clean a string to a proper variable name in python.

    Args:
        text (str): Original string

    Returns:
        str: Corrected string

    .. code-block:: python

        clean_name('32v2 g #Gmw845h$W b53wi ')

    *Output*
        `'_32v2_g__Gmw845h_W_b53wi_'`

    See https://stackoverflow.com/questions/3303312/how-do-i-convert-a-string-to-a-valid-variable-name-in-python
    """
    return re.sub('\W|^(?=\d)', '_', text)


####################################################################################
# Tracebacks

_old_warn = None


def enable_warning_traceback():
    """Show ow traceback on warning."""

    global _old_warn
    _old_warn = warnings.warn

    def warn(*args, **kwargs):
        """Warn user with traceback to warning."""
        tb = traceback.extract_stack()
        _old_warn(*args, **kwargs)
        print("".join(traceback.format_list(tb)[:-1]))

    warnings.warn = warn


def get_traceback():
    """Returns traceback string. Format each frame in the traceback as a
    string.

    Returns:
        str: Traceback string
    """
    trace_back = traceback.extract_stack()
    return "".join(traceback.format_list(trace_back)[:-1])


def print_traceback_easy(start=26):
    """Utility function to print traceback for debug. Will report in series the
    string version of the frames that we are currently in.

    Args:
        start (int): Starting position of the traceback frame.
                     Defaults to 26. Assumes runs from Jupyter notebooks.
                     In general set to zero.
    """
    print(f"\n")
    print('\n'.join(map(repr, traceback.extract_stack()[start:])))
    print('\n')


def log_error_easy(logger: logging.Logger,
                   pre_text='',
                   post_text='',
                   do_print=False):
    """Print log message.

    Args:
        logger (logging.Logger): The logger.
        pre_text (str): Initial text to write.  Defaults to ''.
        post_text (str): End text to write.  Defaults to ''.
        do_print (bool): True to do the printing, False otherwise.  Defaults to False.

    Test use:

    .. code-block:: python

        import traceback
        from qiskit_metal import logger
        from qiskit_metal.toolbox_python.utility_functions import log_error_easy

        def xx():
            exc_info = sys.exc_info()
            try:
                raise TypeError("Oups!")
            except Exception as err:
                try:
                    raise TypeError("Again !?!")
                except:
                    pass

                # exc_type, exc_value, exc_tb = sys.exc_info()
                # error = traceback.format_exception(exc_type, exc_value, exc_tb)
                # logger.error('\\n\\n'+'\\n'.join(error)+'\\n')
                log_error_easy(metal.logger)
        xx()
    """
    exc_type, exc_value, exc_tb = sys.exc_info()
    error = traceback.format_exception(exc_type, exc_value, exc_tb)
    text = f'{pre_text}\n\n' + '\n'.join(error) + f'\n{post_text}'

    logger.error(text)
    if do_print:
        print(text)


#####################################################################################


def monkey_patch(self, func, func_name=None):
    """Monkey patch a method into a class at runtime.

    Use descriptor protocol when adding method as an attribute.

    For a method on a class, when you do a.some_method, python actually does:
         a.some_method.__get__(a, type(a))

    So we're just reproducing that call sequence here explicitly.

    See: https://stackoverflow.com/questions/38485123/monkey-patching-bound-methods-in-python

    Args:
        func (function): function
        func_name (str): name of the function.  Defaults to None.
    """
    func_name = func_name or func.__name__
    setattr(self, func_name, func.__get__(self, self.__class__))
    # what happens if we reload the class or swap in real time?


####################################################################################
# Used to detect and denote potential short segments, when fillet is used.

# # Keep this method until the class QCheckLength will be fully tested.
# def which_vertex_has_potential_fillet_errors(coords: list, a_fillet: float, fillet_comparison_precision: int) -> list:
#     """Iterate through the vertex and check using critea.
#     1. If a start or end segment, is the length smaller than a_fillet.
#     2. If segment in side of LineString, is the lenght smaller than,FILLET_SCALAR times a_fillet.

#     Note, there is a rounding error issues. So when the lenght of the segment is calculated,
#     it is rounded by using fillet_comparison_precision.

#     Args:
#         coords (list): List of tuples in (x,y) format. Each tuple represents a vertex on a LineSegment.

#         a_fillet (float): The radius to fillet a vertex.

#         fillet_comparison_precision (int): There are rounding issues when comparing to (fillet * scalar).
#         Use this when calculating length of line-segment.

#     Returns:
#         list: List of idexes.  Each index corresponds to a vertex in coords, that would not fillet well.
#     """
#     vertex_of_bad = list()
#     len_coords = len(coords)
#     if len_coords <= 1:
#         return vertex_of_bad

#     # When determining the critera to fillet, scale the fillet value by FILLET_SCALAR.
#     FILLET_SCALAR = 2.0
#     scaled_fillet = a_fillet * FILLET_SCALAR

#     for index, xy in enumerate(coords):
#         # Skip the first vertex.
#         if index > 0:
#             xy_previous = coords[index-1]

#             seg_length = np.round(
#                 distance.euclidean(xy_previous, xy), fillet_comparison_precision)

#             # If at first or last segment, use just the fillet value to check, otherwise, use FILLET_SCALAR.
#             # Need to not fillet index-1 to index line segment.
#             if index == 1 or index == len_coords-1:
#                 if seg_length < a_fillet:
#                     vertex_of_bad.extend([index-1, index])
#             else:
#                 if seg_length < scaled_fillet:
#                     vertex_of_bad.extend([index-1, index])

#     # As precaution, remove duplicates from list.
#     vertex_of_bad = list(set(vertex_of_bad))
#     return vertex_of_bad


def toggle_numbers(numbers: list, totlength: int) -> list:
    """
    Given a list of integers called 'numbers', return the toggle of them from zero to totlength - 1.

    Args:
        numbers (list): Integers in the original list, in sorted order.
        totlength (int): Number of elements in complete list. Ex: [0, 1, 2, 3, ..., n - 1] has totlength n.

    Returns:
        list: A sorted list of all integers between 0 and totlength - 1, inclusive, not found in numbers.
    """
    complement = []
    if totlength:
        if not numbers:
            return [i for i in range(totlength)]
        j = 0
        for i in range(totlength):
            if i < numbers[j]:
                complement.append(i)
            else:
                j += 1
                if j >= len(numbers):
                    return complement + [k for k in range(i + 1, totlength)]
    return complement


def bad_fillet_idxs(coords: list,
                    fradius: float,
                    precision: int = 9,
                    isclosed: bool = False) -> list:
    """
    Get list of vertex indices in a linestring (isclosed = False) or polygon (isclosed = True) that cannot be filleted based on
    proximity to neighbors. By default, this list excludes the first and last vertices if the shape is a linestring.

    Args:
        coords (list): Ordered list of tuples of vertex coordinates.
        fradius (float): User-specified fillet radius from QGeometry table.
        precision (int, optional): Digits of precision used for round(). Defaults to 9.
        isclosed (bool, optional): Boolean denoting whether the shape is a linestring or polygon. Defaults to False.

    Returns:
        list: List of indices of vertices too close to their neighbors to be filleted.
    """
    length = len(coords)
    get_dist = Vector.get_distance
    if isclosed:
        return [
            i for i in range(length)
            if min(get_dist(coords[i - 1], coords[i], precision),
                   get_dist(coords[i], coords[(i + 1) %
                                              length], precision)) < 2 * fradius
        ]
    if length < 3:
        return []
    if length == 3:
        return [] if min(get_dist(coords[0], coords[1], precision),
                         get_dist(coords[1], coords[2],
                                  precision)) >= fradius else [1]
    if (get_dist(coords[0], coords[1], precision) < fradius) or (get_dist(
            coords[1], coords[2], precision) < 2 * fradius):
        badlist = [1]
    else:
        badlist = []
    for i in range(2, length - 2):
        if min(get_dist(coords[i - 1], coords[i], precision),
               get_dist(coords[i], coords[i + 1], precision)) < 2 * fradius:
            badlist.append(i)
    if (get_dist(coords[length - 3], coords[length - 2], precision)
            < 2 * fradius) or (get_dist(coords[length - 2], coords[length - 1],
                                        precision) < fradius):
        badlist.append(length - 2)
    return badlist


def good_fillet_idxs(coords: list,
                     fradius: float,
                     precision: int = 9,
                     isclosed: bool = False):
    """
    Get list of vertex indices in a linestring (isclosed = False) or polygon (isclosed = True)
    that can be filleted based on proximity to neighbors.

    Args:
        coords (list): Ordered list of tuples of vertex coordinates.
        fradius (float): User-specified fillet radius from QGeometry table.
        precision (int, optional): Digits of precision used for round(). Defaults to 9.
        isclosed (bool, optional): Boolean denoting whether the shape is a linestring or
            polygon. Defaults to False.

    Returns:
        list: List of indices of vertices that can be filleted.
    """
    if isclosed:
        return toggle_numbers(
            bad_fillet_idxs(coords, fradius, precision, isclosed=True),
            len(coords))
    return toggle_numbers(
        bad_fillet_idxs(coords, fradius, precision, isclosed=False),
        len(coords))[1:-1]


def get_range_of_vertex_to_not_fillet(coords: list,
                                      fradius: float,
                                      precision: int = 9,
                                      add_endpoints: bool = True) -> list:
    """Provide a list of tuples for a list of integers that correspond to
    coords. Each tuple corresponds to a range of indexes within coords.  A
    range denotes vertexes that are too short to be fillet'd.

    If the range is just one point, meaning, not a segment, the tuple will contain
    the same index for start and end.

    Args:
        coords (list): Ordered list of tuples of vertex coordinates.
        fradius (float): User-specified fillet radius from QGeometry table.
        precision (int, optional): Digits of precision used for round(). Defaults to 9.
        add_endpoints (bool): Default is True.  If the second to endpoint is in list,
            add the endpoint to list.  Used for GDS, not add_qgeometry.

    Returns:
        list: A compressed list of tuples.  So, it combines adjacent vertexes into a longer one.
    """
    length = len(coords)

    # isclosed=False is for LineString
    unique_vertex = bad_fillet_idxs(coords, fradius, precision, isclosed=False)

    if add_endpoints:
        # The endpoints of LineString are never fillet'd. If the second vertex or second to last vertex
        # should not be fillet's, then don't fillet the endpoints.  This is used for warning for add_qgeometry.
        # Also used in QGDSRenderer when breaking the LineString.
        if (1 in unique_vertex) and (0 not in unique_vertex):
            unique_vertex.append(0)

        # second to last vertex in unique_vertex
        if (length - 2 in unique_vertex) and (length - 1 not in unique_vertex):
            unique_vertex.append(length - 1)

    compressed_vertex = compress_vertex_list(unique_vertex)

    return compressed_vertex


def compress_vertex_list(individual_vertex: list) -> list:
    """Given a list of vertices that should not be fillet'd, search for a range
    and make them one compressed list. If the vertex is a point and not a line
    segment, the returned tuple's start and end are the same index.

    Args:
        individual_vertex (list): List of UNIQUE ints.  Each int refers to an index of a LineString.

    Returns:
        list: A compressed list of tuples.  So, it combines adjacent vertices into a longer one.
    """
    reduced_idx = list()

    sorted_vertex = sorted(individual_vertex)
    len_vertex = len(sorted_vertex)

    if len_vertex > 0:
        # initialzie to unrealistic number.
        start = -1
        end = -1
        size_of_range = 0

        for index, item in enumerate(sorted_vertex):
            if index == 0:
                start = item
                end = item
            else:
                if item == end + 1:
                    end = item
                    size_of_range += 1
                else:
                    if size_of_range == 0:
                        # Only one vertex in range.
                        reduced_idx.append((start, end))
                        start = item
                        end = item
                    else:
                        # Two or more vertexes in range.
                        reduced_idx.append((start, end))
                        size_of_range = 0
                        start = item
                        end = item

            if index == len_vertex - 1:
                if size_of_range == 0:
                    reduced_idx.append((start, end))
                else:
                    reduced_idx.append((start, end))
        return reduced_idx
    else:
        return reduced_idx


#######################################################################################
# File checking


def can_write_to_path_with_warning(file: str) -> int:
    """Check if can write file.

    Args:
        file (str): Has the path and/or just the file name.

    Returns:
        int: 1 if access is allowed. Else returns 0, if access not given.
    """
    a_logger = logger
    # If need to use lib pathlib.
    directory_name = os.path.dirname(os.path.abspath(file))
    if os.access(directory_name, os.W_OK):
        return 1
    else:
        a_logger.warning(f'Not able to write to directory.'
                         f'File:"{file}" not written.'
                         f' Checked directory:"{directory_name}".')
        return 0


def can_write_to_path(file: str) -> Tuple[int, str]:
    """Check to see if path exists and file can be written.

    Args:
        file (str): Has the path and/or just the file name.

    Returns:
        Tuple[int, str]:
        int: 1 if access is allowed. Else returns 0, if access not given.
        str: Full path and file which was searched for.
    """
    # If need to use lib pathlib.
    directory_name = os.path.dirname(os.path.abspath(file))
    if os.access(directory_name, os.W_OK):
        return 1, directory_name
    else:
        return 0, directory_name


#####################################################################################
# function signature


def check_all_required_args_provided(func: Callable,
                                     args: Dict,
                                     raise_exception: bool = True) -> List:
    """Check all required arguments of func are provided by args

    Args:
        func (Callable): function callable to check
        args (Dict): dictionary of provided arguments
        raise_exception (bool=True): raise InputError exception if missing args found

    Returns:
        List: list of missing argument names

    """
    provided_args = list(args.keys())
    missing_args = []
    for param in inspect.signature(func).parameters.values():
        if param.name in provided_args or param.name == 'self':
            continue
        if param.default is inspect._empty and param.name not in provided_args:
            missing_args.append(param.name)

    if missing_args and raise_exception:
        out = ', '.join(missing_args)
        raise InputError(
            f'Missing values for arguments {out} for {func.__qualname__}')
    return missing_args


def get_all_args(func: Callable) -> List:
    """get all the parameters of a function

    Args:
        func (Callable): function from which to get parameters

    Returns:
        List: list of parameters of the function
    """
    args = []
    for param in inspect.signature(func).parameters.values():
        if param.name == 'self':
            continue
        args.append(param.name)
    return args


#####################################################################################
# Used for each row in qgeometry table to clean the name entered by user.


def get_clean_name(name: str) -> str:
    """Create a valid variable name from the given one by removing having it
    begin with a letter or underscore followed by an unlimited string of
    letters, numbers, and underscores.

    Args:
        name (str): Initial, possibly unusable, string to be modified.

    Returns:
        str: Variable name consistent with Python naming conventions.
    """
    # Remove invalid characters
    name = re.sub("[^0-9a-zA-Z_]", "", name)
    # Remove leading characters until we find a letter or underscore
    name = re.sub("^[^a-zA-Z_]+", "", name)
    return name



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""QRendererGui Base."""

from ...designs import QDesign, is_design
from ...qgeometries import QGeometryTables

from . import QRenderer

__all__ = ['QRendererGui']


class QRendererGui(QRenderer):
    """Abstract base class for the GUI rendering.

    Extends `QRenderer`. An interface class.
    """

    name = 'guibase'  # overwrite this!
    """Name"""

    def __init__(self, gui, design: QDesign, initiate=True, fig=None, ax=None):
        """
        Args:
            gui (MetalGUI): the GUI.
            design (QDesign): The design.
            initiate (bool): True to initiate the renderer.  Defaults to True.
            fig (figure): current figure.  Defaults to None.
            ax (ax): current ax.  Defaults to None.
        """
        super().__init__(design=design, initiate=initiate)

        self.gui = gui
        self.fig = None  # current figure
        self.ax = None  # current ax

    def set_fig(self, fig):
        """Set the given figure.

        Args:
            fig (figure): Figure to set
        """
        self.fig = fig

    def set_ax(self, ax):
        """Set the given ax.

        Args:
            ax (ax): ax to set
        """
        self.ax = ax

    def setup_fig(self, fig):
        """Setup the given figure.

        Args:
            fig (figure): figure to setup

        Raises:
            NotImplementedError: Function not written yet
        """
        raise NotImplementedError()

    def style_axis(self, ax):
        """Style the axis.

        Args:
            ax (ax): ax to style

        Raises:
            NotImplementedError: Function not written yet
        """
        raise NotImplementedError()

    def render_design(self, selection=None):
        """Render the design.

        Args:
            selection (selection): Not used
        """
        # TOOD: handle selection
        for _, component in self.design.components.items():
            self.render_component(component)

    def render_component(self, component):
        """Render the given component.

        Args:
            component (QComponent): the component

        Raises:
            NotImplementedError: Function not written yet
        """
        raise NotImplementedError()

    def render_shapely(self):
        """Render shapely.

        Raises:
            NotImplementedError: Function not written yet
        """
        raise NotImplementedError()

    def render_connectors(self):
        """Render connectors.

        Raises:
            NotImplementedError: Function not written yet
        """
        raise NotImplementedError()

    def clear_axis(self):
        """Clear the axis.

        Raises:
            NotImplementedError: Function not written yet
        """
        raise NotImplementedError()

    def clear_figure(self):
        """Clear  the figure.

        Raises:
            NotImplementedError: Function not written yet
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=protected-access
# pylint: disable-msg=pointless-statement
# pylint: disable-msg=too-many-public-methods
# pylint: disable-msg=broad-except
# pylint: disable-msg=invalid-name
# pylint: disable-msg=import-error
"""Qiskit Metal unit tests analyses functionality."""

import unittest
import pandas as pd

from qiskit_metal.designs.design_base import QDesign
from qiskit_metal.designs.design_planar import DesignPlanar
from qiskit_metal.designs.interface_components import Components
from qiskit_metal.designs.net_info import QNet
from qiskit_metal.qlibrary.core import QComponent
from qiskit_metal.qlibrary.qubits.transmon_pocket import TransmonPocket
from qiskit_metal.tests.assertions import AssertionsMixin

from qiskit_metal.qlibrary.lumped.resonator_coil_rect import ResonatorCoilRect


class TestDesign(unittest.TestCase, AssertionsMixin):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    def test_design_instantiate_qdesign(self):
        """Test the instantiation of QDesign."""
        try:
            QDesign()
        except Exception:
            self.fail("QDesign failed")

        try:
            QDesign(metadata={})
        except Exception:
            self.fail("QDesign(metadata={}) failed")

        try:
            QDesign(metadata={}, overwrite_enabled=True)
        except Exception:
            self.fail("QDesign(metadata={}, overwrite_enabled=True) failed")

    def test_design_instantiate_design_planar(self):
        """Test the instantiation of DesignPlanar."""
        try:
            DesignPlanar()
        except Exception:
            self.fail("DesignPlanar failed")

        try:
            DesignPlanar(metadata={})
        except Exception:
            self.fail("DesignPlanar(metadata={}) failed")

        try:
            DesignPlanar(metadata={}, overwrite_enabled=True)
        except Exception:
            self.fail(
                "DesignPlanar(metadata={}, overwrite_enabled=True) failed")

    def test_design_instantiate_design_components(self):
        """Test the instantiation of Components."""
        design = QDesign()
        try:
            Components(design)
        except Exception:
            self.fail("Components(design) failed")

    def test_design_instantiate_qnet(self):
        """Test the instantiation of QNet."""
        try:
            QNet()
        except Exception:
            self.fail("QNet failed")

    def test_design_get_chip_layer(self):
        """Test getting chip size in design_base.py."""
        design = QDesign(metadata={})
        self.assertEqual(design.get_chip_layer('main'), 0)

    def test_design_rename_variable(self):
        """Test rename_variable in QDesign class"""
        design = QDesign(metadata={})
        design.rename_variable('cpw_gap', 'new_name')
        my_keys = design._variables.keys()

        self.assertEqual(len(my_keys), 2)
        self.assertTrue('cpw_width' in my_keys)
        self.assertTrue('new_name' in my_keys)

    def test_design_update_metadata(self):
        """Test the update of metadata in design_base.py."""
        design = DesignPlanar(metadata={})
        data = design.metadata

        design.update_metadata({'new-key': 'new-value'})
        design.update_metadata({'time_created': '07/23/2020, 17:00:00'})
        data = design.metadata

        self.assertEqual(len(data), 4)
        self.assertEqual(data['design_name'], 'my_design')
        self.assertEqual(data['notes'], '')
        self.assertEqual(data['time_created'], '07/23/2020, 17:00:00')
        self.assertEqual(data['new-key'], 'new-value')

    def test_design_get_chip_size(self):
        """Test getting chip size in design_base.py."""
        design = DesignPlanar(metadata={})

        main_expected = {
            'center_x': '0.0mm',
            'center_y': '0.0mm',
            'center_z': '0.0mm',
            'size_x': '9mm',
            'size_y': '6mm',
            'size_z': '-750um',
            'sample_holder_top': '890um',
            'sample_holder_bottom': '1650um'
        }

        self.assertEqual(design.get_chip_size(), main_expected)
        self.assertEqual(design.get_chip_size('main'), main_expected)
        self.assertEqual(design.get_chip_size('dead'), {})

    def test_design_get_chip_z(self):
        """Test getting chip z in design_base.py."""
        design = DesignPlanar(metadata={})

        actual = design.get_chip_z()
        self.assertEqual(actual, '0.0mm')

    def test_design_rename_variables(self):
        """Test variable renaming in design_base.py."""
        design = DesignPlanar(metadata={})
        design.rename_variable('cpw_gap', 'new-name')

        with self.assertRaises(ValueError):
            design.rename_variable('i-do-not-exist', 'uh-oh')

        self.assertEqual('new-name' in design.variables.keys(), True)
        self.assertEqual('cpw_gap' in design.variables.keys(), False)

    def test_design_rename_component(self):
        """Test renaming component in design_base.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)

        design.rename_component('my_name-1', 'new-name')

        self.assertEqual('new-name' in design.name_to_id, True)
        self.assertEqual('my_name-1' in design.name_to_id, False)
        self.assertEqual('my_name-2' in design.name_to_id, True)

    def test_design_default_component_name(self):
        """Test automatic naming of components."""
        design = DesignPlanar(metadata={})

        ResonatorCoilRect(design, make=False)
        self.assertEqual('res_1' in design.components, True)
        ResonatorCoilRect(design, make=False)
        self.assertEqual('res_2' in design.components, True)

        # Manually add the next automatic name to check it doesn't get repeated
        ResonatorCoilRect(design, 'res_3', make=False)
        ResonatorCoilRect(design, make=False)
        self.assertEqual('res_3' in design.components, True)
        self.assertEqual('res_4' in design.components, True)

        # Add a different component
        TransmonPocket(design, make=False)
        self.assertEqual('Q_1' in design.components, False)

        # Add a component with no predefined prefix
        QComponent(design, make=False)
        self.assertEqual('QComponent_1' in design.components, True)

    def test_design_delete_component(self):
        """Test deleteing a component in design_base.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)

        design.delete_component('my_name-1')
        self.assertEqual('my_name-1' in design.name_to_id, False)
        self.assertEqual('my_name-2' in design.name_to_id, True)

    def test_design_delete_all_components(self):
        """Test deleteing a component in design_base.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)

        design.delete_all_components()
        self.assertEqual('my_name-1' in design.name_to_id, False)
        self.assertEqual('my_name-2' in design.name_to_id, False)

    def test_design_get_and_set_design_name(self):
        """Test getting the design name in design_base.py."""
        design = DesignPlanar(metadata={})
        self.assertEqual(design.get_design_name(), 'my_design')

        design.set_design_name('new-name')
        self.assertEqual(design.get_design_name(), 'new-name')

    def test_design_get_units(self):
        """Test get units in design_base.py."""
        design = DesignPlanar(metadata={})
        self.assertEqual(design.get_units(), 'mm')

    def test_design_get_new_qcomponent_name_id(self):
        """Test _get_new_qcomponent_name_id in design_base.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, make=False)
        QComponent(design, make=False)

        self.assertEqual(design._get_new_qcomponent_name_id('QComponent'), 3)
        self.assertEqual(design._get_new_qcomponent_name_id('Not-there'), 1)

    def test_design_planar_get_x_y_for_chip(self):
        """Test get_x_y_for_chip in design_planar.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, make=False)
        QComponent(design, make=False)

        expected = ((-4.5, -3.0, 4.5, 3.0), 0)
        actual = design.get_x_y_for_chip('main')

        self.assertEqual(len(expected), len(actual))
        self.assertEqual(len(expected[0]), len(actual[0]))
        for i in range(4):
            self.assertEqual(expected[0][i], actual[0][i])

        self.assertEqual(expected[1], actual[1])

    def test_design_interface_components_get_list_ints(self):
        """Test geting the list ints in interface_components.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)
        components = Components(design)

        self.assertListEqual(components.get_list_ints(['my_name-1']), [1])
        self.assertListEqual(
            components.get_list_ints(['my_name-1', 'my_name-2']), [1, 2])
        self.assertListEqual(
            components.get_list_ints(['my_name-1', 'my_name-2', 'nope']),
            [1, 2, 0])

    def test_design_interface_components_find_id(self):
        """Test finding the id from the name in interface_components.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)
        components = Components(design)

        self.assertEqual(components.find_id('my_name-1'), 1)
        self.assertEqual(components.find_id('my_name-2'), 2)
        self.assertEqual(components.find_id('my_name-3'), 0)

    def test_design_qnet_add_pins_to_table(self):
        """Test add_pins_to_table in net_info.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)
        Components(design)
        qnet = QNet()

        qnet.add_pins_to_table(1, 'my_name-1', 2, 'my_name-2')
        df = qnet.net_info

        data = {
            'net_id': [1, 1],
            'component_id': [1, 2],
            'pin_name': ['my_name-1', 'my_name-2']
        }
        df_expected = pd.DataFrame(data, index=[0, 1])

        self.assertEqual(len(df), len(df_expected))
        data_points = df_expected['net_id'].size
        for i in range(data_points):
            for j in ['net_id', 'component_id', 'pin_name']:
                self.assertEqual(df_expected[j][i], df[j][i])

    def test_design_qnet_delete_net_id(self):
        """Test delete a given net id in net_info.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)
        Components(design)
        qnet = QNet()

        net_id = qnet.add_pins_to_table(1, 'my_name-1', 2, 'my_name-2')
        qnet.delete_net_id(net_id)
        df = qnet.net_info
        self.assertEqual(df.empty, True)

    def test_design_qnet_delete_all_pins_for_component_id(self):
        """Test delete all pins for a given component id in net_info.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)
        QComponent(design, 'my_name-3', make=False)
        QComponent(design, 'my_name-4', make=False)
        Components(design)
        qnet = QNet()

        qnet.add_pins_to_table(1, 'my_name-1', 2, 'my_name-2')
        qnet.add_pins_to_table(3, 'my_name-3', 4, 'my_name-4')
        qnet.delete_all_pins_for_component(2)
        df = qnet.net_info

        data = {
            'net_id': [2, 2],
            'component_id': [3, 4],
            'pin_name': ['my_name-3', 'my_name-4']
        }
        df_expected = pd.DataFrame(data, index=[2, 3])

        self.assertEqual(len(df), len(df_expected))
        for i in [2, 3]:
            for j in ['net_id', 'component_id', 'pin_name']:
                self.assertEqual(df_expected[j][i], df[j][i])

    def test_design_qnet_get_components_and_pins_for_netid(self):
        """Test get_components_Sand_pins_for_netid in net_info.py."""
        design = DesignPlanar(metadata={})
        QComponent(design, 'my_name-1', make=False)
        QComponent(design, 'my_name-2', make=False)
        QComponent(design, 'my_name-3', make=False)
        QComponent(design, 'my_name-4', make=False)
        Components(design)
        qnet = QNet()

        net_id_1 = qnet.add_pins_to_table(1, 'my_name-1', 2, 'my_name-2')
        net_id_2 = qnet.add_pins_to_table(3, 'my_name-3', 4, 'my_name-4')

        data = {
            'net_id': [1, 1],
            'component_id': [1, 2],
            'pin_name': ['my_name-1', 'my_name-2']
        }
        df_expected_1 = pd.DataFrame(data, index=[0, 1])

        data = {
            'net_id': [2, 2],
            'component_id': [3, 4],
            'pin_name': ['my_name-3', 'my_name-4']
        }
        df_expected_2 = pd.DataFrame(data, index=[2, 3])

        df = qnet.get_components_and_pins_for_netid(net_id_1)
        self.assertEqual(len(df), len(df_expected_1))
        data_points = df_expected_1['net_id'].size
        for i in range(data_points):
            for j in ['net_id', 'component_id', 'pin_name']:
                self.assertEqual(df_expected_1[j][i], df[j][i])

        df = qnet.get_components_and_pins_for_netid(net_id_2)
        self.assertEqual(len(df), len(df_expected_2))
        for i in [2, 3]:
            for j in ['net_id', 'component_id', 'pin_name']:
                self.assertEqual(df_expected_2[j][i], df[j][i])

    def test_design_copy_qcomponent(self):
        """Test the functionlaity of copy_qcomponent in design_base.py."""
        design = DesignPlanar()
        design.overwrite_enabled = True
        q_1 = TransmonPocket(design, 'Q1')
        q1_copy = design.copy_qcomponent(q_1, 'Q1_copy')

        self.assertEqual(q_1.name, 'Q1')
        self.assertEqual(q1_copy.name, 'Q1_copy')
        self.assertEqual(q_1.id, 1)
        self.assertEqual(q1_copy.id, 2)
        self.assertEqual(q_1.class_name, q1_copy.class_name)
        self.assertEqual(q_1.options['pos_x'], q1_copy.options['pos_x'])

        q1_copy.options['pos_x'] = '1.0mm'
        self.assertEqual(q_1.options['pos_x'], '0.0um')
        self.assertEqual(q1_copy.options['pos_x'], '1.0mm')

    def test_design_copy_multiple_qcomponents(self):
        """Test the functionality of copy_multiple_qcomponents in
        design_base.py."""
        design = DesignPlanar()
        design.overwrite_enabled = True
        q_1 = TransmonPocket(design, 'Q1')
        q_2 = TransmonPocket(design, 'Q2')
        q_3 = TransmonPocket(design, 'Q3')

        design.copy_multiple_qcomponents(
            [q_1, q_2, q_3], ['q1_copy', 'q2_copy', 'q3_copy'],
            [dict(pos_y='-1.0mm'),
             dict(pos_y='-2.0mm'),
             dict(pos_y='-3.0mm')])

        q1_copy = design._components[4]
        q2_copy = design._components[5]
        q3_copy = design._components[6]

        self.assertEqual(q_1.name, 'Q1')
        self.assertEqual(q_2.name, 'Q2')
        self.assertEqual(q_3.name, 'Q3')
        self.assertEqual(q1_copy.name, 'q1_copy')
        self.assertEqual(q2_copy.name, 'q2_copy')
        self.assertEqual(q3_copy.name, 'q3_copy')
        self.assertEqual(q_1.id, 1)
        self.assertEqual(q_2.id, 2)
        self.assertEqual(q_3.id, 3)
        self.assertEqual(q1_copy.id, 4)
        self.assertEqual(q2_copy.id, 5)
        self.assertEqual(q3_copy.id, 6)

        self.assertEqual(q_1.class_name, q1_copy.class_name)
        self.assertEqual(q_2.class_name, q2_copy.class_name)
        self.assertEqual(q_3.class_name, q3_copy.class_name)
        self.assertEqual(q_1.options['pos_x'], q1_copy.options['pos_x'])
        self.assertEqual(q_2.options['pos_x'], q2_copy.options['pos_x'])
        self.assertEqual(q_3.options['pos_x'], q3_copy.options['pos_x'])

        self.assertEqual(q1_copy.options['pos_y'], '-1.0mm')
        self.assertEqual(q2_copy.options['pos_y'], '-2.0mm')
        self.assertEqual(q3_copy.options['pos_y'], '-3.0mm')

    def test_design_connect_pins(self):
        """Test connect_pins functionality in design_base.py."""
        design = DesignPlanar()
        design.overwrite_enabled = True

        TransmonPocket(design, 'Q1')
        TransmonPocket(design, 'Q2')

        design.connect_pins(1, 'p1', 2, 'p2')
        pf = design._qnet._net_info

        self.assertFalse(pf.empty)
        self.assertEqual(pf['net_id'][0], 1)
        self.assertEqual(pf['net_id'][1], 1)
        self.assertEqual(pf['component_id'][0], 1)
        self.assertEqual(pf['component_id'][1], 2)
        self.assertEqual(pf['pin_name'][0], 'p1')
        self.assertEqual(pf['pin_name'][1], 'p2')

    def test_design_delete_all_pins(self):
        """Test delete_all_pins functionality in design_base.py."""
        design = DesignPlanar()
        design.overwrite_enabled = True

        TransmonPocket(design, 'Q1')
        TransmonPocket(design, 'Q2')

        design.connect_pins(1, 'p1', 2, 'p2')
        pf = design._qnet._net_info
        self.assertFalse(pf.empty)

        design.delete_all_pins()
        pf = design._qnet._net_info
        self.assertTrue(pf.empty)

    def test_design_all_component_names_id(self):
        """Test all_component_names_id functionality in design_base.py."""
        design = DesignPlanar()
        design.overwrite_enabled = True

        TransmonPocket(design, 'Q1')
        TransmonPocket(design, 'Q2')

        expected = [('Q1', 1), ('Q2', 2)]
        actual = design.all_component_names_id()

        for i in range(2):
            self.assertEqual(expected[i], actual[i])

    def test_design_get_list_of_tables_in_metadata(self):
        """Tests the get_list_of_tables_in_metadata function in design_base.py
        by exeucting get_table_values_form_renderers in a component."""
        design = DesignPlanar()
        q1 = TransmonPocket(design, 'Q1')

        result = q1._get_table_values_from_renderers(design)

        self.assertEqual(len(result), 17)
        self.assertEqual(result['hfss_inductance'], '10nH')
        self.assertEqual(result['hfss_capacitance'], 0)
        self.assertEqual(result['hfss_resistance'], 0)
        self.assertAlmostEqual(result['hfss_mesh_kw_jj'], 7e-06, places=6)
        self.assertEqual(result['q3d_inductance'], '10nH')
        self.assertEqual(result['q3d_capacitance'], 0)
        self.assertEqual(result['q3d_resistance'], 0)
        self.assertAlmostEqual(result['q3d_mesh_kw_jj'], 7e-06, places=6)
        self.assertEqual(result['gds_cell_name'], 'my_other_junction')
        self.assertEqual(result['hfss_wire_bonds'], False)
        self.assertEqual(result['q3d_wire_bonds'], False)
        self.assertEqual(result['aedt_hfss_inductance'], 10e-9)
        self.assertEqual(result['aedt_hfss_capacitance'], 0)


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Main module that handles the elements window inside the main window."""

from typing import TYPE_CHECKING

from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QAbstractTableModel, QModelIndex
from PySide6.QtWidgets import QMainWindow

from .elements_ui import Ui_ElementsWindow

if TYPE_CHECKING:
    # https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports
    from .main_window import MetalGUI, QMainWindowExtension


class ElementsWindow(QMainWindow):
    """This is just a handler (container) for the UI; it a child object of the
    main gui.

    Extends the `QMainWindow` class.

    PySide6 Signal / Slots Extensions:
        The UI can call up to this class to execute button clicks for instance
        Extensiosn in qt designer on signals/slots are linked to this class
    """

    def __init__(self, gui: 'MetalGUI', parent_window: 'QMainWindowExtension'):
        """
        Args:
            gui (MetalGUI): The GUI
            parent_window (QMainWindowExtension): Parent window
        """
        # Q Main Window
        super().__init__(parent_window)

        # Parent GUI related
        self.gui = gui
        self.logger = gui.logger
        self.statusbar_label = gui.statusbar_label

        # UI
        self.ui = Ui_ElementsWindow()
        self.ui.setupUi(self)

        self.statusBar().hide()

        self.model = ElementTableModel(gui, self)
        self.ui.tableElements.setModel(self.model)

    @property
    def design(self):
        """Returns the design."""
        return self.gui.design

    def populate_combo_element(self):
        """Populate the combo elements."""
        for table_type in self.design.qgeometry.tables.keys():
            if self.ui.combo_element_type.findText(
                    table_type) == -1:  # not in combo box, add it
                self.ui.combo_element_type.addItem(
                    str(
                        QtWidgets.QApplication.translate(
                            "ElementsWindow", table_type, None, -1)))

    def combo_element_type(self, new_type: str):
        """Change to the given type.

        Args:
            new_type (str): Type to change to
        """
        self.logger.info(f'Changed element table type to: {new_type}')
        self.model.set_type(new_type)

    def force_refresh(self):
        """Force a refresh."""
        self.model.refresh()
        self.populate_combo_element()


class ElementTableModel(QAbstractTableModel):
    """MVC ElementTableModel class See
    https://doc.qt.io/qt-5/qabstracttablemodel.html.

    The class extends the `QAbstractTableModel` class.

    Can be accessed with:
        .. code-block:: python

            t = gui.elements_win.tableElements
            model = t.model()
            index = model.index(1,0)
            model.data(index)
    """
    __timer_interval = 500  # ms

    def __init__(self, gui, parent=None, element_type='poly'):
        """
        Args:
            gui (MetalGUI): The GUI
            parent (QMainWindowExtension): Parent window.  Defaults to None.
            element_type (str): The element type.  Defaults to 'poly'.
        """
        super().__init__(parent=parent)
        self.logger = gui.logger
        self.gui = gui
        self._row_count = -1
        self.type = element_type

        self._create_timer()

    @property
    def design(self):
        """Returns the design."""
        return self.gui.design

    @property
    def qgeometry(self):
        """Returns the qgeometry."""
        if self.design:
            return self.design.qgeometry

    @property
    def tables(self):
        """Returns all the tables."""
        if self.design:
            return self.design.qgeometry.tables

    @property
    def table(self):
        """Returns all the tables of the type specified in the constructor."""
        if self.design:
            return self.design.qgeometry.tables[self.type]

    def _create_timer(self):
        """Refresh the model number of rows, etc."""
        self._timer = QtCore.QTimer(self)
        self._timer.start(self.__timer_interval)
        self._timer.timeout.connect(self.refresh_auto)

    def set_type(self, element_type: str):
        """Set the type.

        Args:
            element_type (str): Element type to set to
        """
        self.type = element_type
        self.refresh()

    def refresh(self):
        """Force refresh.

        Completely rebuild the model.
        """
        self.beginResetModel()
        try:
            # parent_index = self.createIndex(0, 0, self.root)
            self._row_count = self.rowCount(None)
        finally:
            self.endResetModel()

    def refresh_auto(self):
        """Update row count etc."""
        new_count = self.rowCount()

        # if the number of rows have changed
        if self._row_count != new_count:
            #self.logger.info('Number of components changed')

            # Wrap the reset logic in beginResetModel and endResetModel
            self.beginResetModel()
            try:

                # When a model is reset it should be considered that all
                # information previously retrieved from it is invalid.
                # This includes but is not limited to the rowCount() and
                # columnCount(), flags(), data retrieved through data(), and roleNames().
                # This will loose the current selection.
                # self.modelReset.emit()

                self._row_count = new_count
            finally:
                self.endResetModel()

    def rowCount(self, parent: QModelIndex = None) -> int:
        """Counts all the rows.

        Args:
            parent (QModelIndex): Unused.  Defaults to None.

        Returns:
            int: The number of rows
        """
        if parent is None:
            parent = QModelIndex()
        if self.table is None:
            return 0
        return self.table.shape[0]

    def columnCount(self, parent: QModelIndex = None):  #pylint: disable=unused-argument
        """Counts all the columns.

        Args:
            parent (QModelIndex): Unused.  Defaults to None.

        Returns:
            int: The number of columns
        """
        if self.table is None:
            return 0
        return self.table.shape[1]

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        """Set the headers to be displayed.

        Args:
            section (int): Section number
            orientation (Qt orientation): Section orientation
            role (Qt display role): Display role.  Defaults to DisplayRole.

        Returns:
            str: The header data, or None if not found
        """

        if (role != QtCore.Qt.DisplayRole) or (self.table is None):
            return None

        if orientation == QtCore.Qt.Horizontal:
            if section < self.columnCount():
                return str(self.table.columns[section])

    def flags(self, index: QModelIndex):
        """Set the item flags at the given index. Seems like we're implementing
        this function just to see how it's done, as we manually adjust each
        tableView to have NoEditTriggers.

        Args:
            index (QModelIndex): The index

        Returns:
            Qt flags: Flags from Qt
        """
        # https://doc.qt.io/qt-5/qt.html#ItemFlag-enum

        if not index.isValid():
            return QtCore.Qt.ItemIsEnabled

        return QtCore.Qt.ItemFlags(
            QAbstractTableModel.flags(self, index) |
            QtCore.Qt.ItemIsSelectable)  # ItemIsEditable

    def data(self, index: QModelIndex, role=QtCore.Qt.DisplayRole):
        """Depending on the index and role given, return data. If not returning
        data, return None (PySide equivalent of QT's "invalid QVariant").

        Returns:
            str: Data related to the given index and role
        """

        if not index.isValid():
            return
        # if not 0 <= index.row() < self.rowCount():
        #    return None

        if self.table is None:
            return

        if role == QtCore.Qt.DisplayRole:
            row = index.row()
            column = index.column()
            # First column (component id) members, are ints so
            # they should sort as numbers instead of strings.
            if column == 0:
                return self.table.iloc[row, column]
            else:
                return str(self.table.iloc[row, column])



# -*- coding: utf-8 -*-
# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Lumped Resonator, as shown in Phys. Rev. Appl. 10, 034050 (2018).
"""
from math import sin, cos
import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import QComponent


class ResonatorLumped(QComponent):
    '''
    The base ResonatorLumped class
    Inherits the QComponent class

    .. TODO: add resonator_lumped.png once asset is available for docs

    .. meta::
        :description: Lumped Resonator

    Default Options:
        * pos_x: '0um' -- x-coordinate of the bottom center of the resonator
        * pos_y: '0um' -- y-coordinate of the bottom center of the resonator
        * orientation: '0' -- angle of rotation of the resonator
        * box_width: '5mm' -- the width of the rectangular resonator structure
        * box_height: '5mm' -- the height of the rectangular resonator structure
        * perimeter_thicknes: '0.01mm' -- the width of the resonator along the perimeter
        * res_width: '0.01m' -- the width of the resonator
        * initial: '0.1m' -- the length of the lower straight resonator segment
        * n_turns: '10' -- the number of turns in the curved resonator segment
        * turn_radius: '0.1mm' - the radius of the turns in the curved resonator segment
        * inner_space: '0.1mm' - the space between the curved resonator
        * outer_space: '0.1mm' -- the space between rectangular and curved resonator turns
        * final: '2.5mm' -- the length of the upper straight segment of the resonator
        * break_width: '0.2mm' -- the width of the "break" in the top of the rectangular resonator
        * layer: '1' -- the layer of the component
    '''

    default_options = Dict(
        pos_x='0 um',
        pos_y='0 um',
        orientation='0',
        box_width='5mm',
        box_height='5mm',
        perimeter_thickness='0.01mm',
        res_width='0.01mm',
        initial='0.1mm',
        n_turns='10',
        turn_radius='0.1mm',
        inner_space='0.1mm',
        outer_space='0.1mm',
        final='2.5mm',
        break_width='0.2mm',
        layer='1',
    )

    ##########################
    def make(self):
        """Builds the component."""
        """Convert self.options into QGeometry."""
        p = self.parse_options()  # Parse the string options into numbers

        # draw the perimeter
        box_out = draw.rectangle(p.box_width, p.box_height, 0.0, 0.0)
        box_in = draw.rectangle(p.box_width - 2.0 * p.perimeter_thickness,
                                p.box_height - 2.0 * p.perimeter_thickness, 0.0,
                                0.0)
        perimeter = draw.subtract(box_out, box_in)
        perimeter = draw.translate(perimeter, xoff=0.0, yoff=0.5 * p.box_height)
        break_rect = draw.rectangle(p.break_width, p.break_width, 0.0,
                                    p.box_height)
        perimeter = draw.subtract(perimeter, break_rect)
        # start the resonator
        initial = draw.LineString([(0.0, p.perimeter_thickness),
                                   (0.0, p.perimeter_thickness + p.initial)])
        # Draw the first quarter turn
        centerx, centery = -1.0 * p.turn_radius, p.perimeter_thickness + p.initial
        radius = p.turn_radius
        start_angle, end_angle = 0, 90  # In degrees
        numsegments = 1000
        # The coordinates of the arc
        theta = np.radians(np.linspace(start_angle, end_angle, numsegments))
        x = centerx + radius * np.cos(theta)
        y = centery + radius * np.sin(theta)
        arc = draw.LineString(np.column_stack([x, y]))
        # Draw the first half-turn (left); this gets repeated
        centerx2 = -0.5 * p.box_width + p.outer_space + p.turn_radius
        centery2 = p.perimeter_thickness + p.initial + 2.0 * p.turn_radius
        radius2 = p.turn_radius
        start_angle2, end_angle2 = 270, 90  # In degrees
        numsegments2 = 1000
        # The coordinates of the arc
        theta2 = np.radians(np.linspace(start_angle2, end_angle2, numsegments2))
        x2 = centerx2 + radius2 * np.cos(theta2)
        y2 = centery2 + radius2 * np.sin(theta2)
        arc_left_1 = draw.LineString(np.column_stack([x2, y2]))
        # Draw the first half-turn (right); this gets repeated
        centerx3 = 0.5 * p.box_width - p.outer_space - p.turn_radius
        centery3 = p.perimeter_thickness + p.initial + 4.0 * p.turn_radius
        radius3 = p.turn_radius
        start_angle3, end_angle3 = 90, -90  # In degrees
        numsegments3 = 1000
        # The coordinates of the arc
        theta3 = np.radians(np.linspace(start_angle3, end_angle3, numsegments3))
        x3 = centerx3 + radius3 * np.cos(theta3)
        y3 = centery3 + radius3 * np.sin(theta3)
        arc_right_1 = draw.LineString(np.column_stack([x3, y3]))
        # bottom half-line
        line1 = draw.LineString([
            (-1.0 * p.turn_radius - 0.0 * p.res_width,
             p.perimeter_thickness + p.initial + p.turn_radius),
            (-0.5 * p.box_width + p.outer_space + p.turn_radius,
             p.perimeter_thickness + p.initial + p.turn_radius)
        ])
        # bottom full-line (this one gets repeated)
        line2 = draw.LineString([
            (-0.5 * p.box_width + p.outer_space + p.turn_radius,
             p.initial + 3.0 * p.turn_radius + p.res_width),
            (0.5 * p.box_width - p.outer_space - p.turn_radius,
             p.initial + 3.0 * p.turn_radius + p.res_width)
        ])
        # repeat the full fline
        line3 = draw.translate(line2, 0.0, 2.0 * p.turn_radius)
        line4 = draw.translate(line3, 0.0, 2.0 * p.turn_radius)
        line5 = draw.translate(line4, 0.0, 2.0 * p.turn_radius)
        line6 = draw.translate(line5, 0.0, 2.0 * p.turn_radius)
        line7 = draw.translate(line6, 0.0, 2.0 * p.turn_radius)
        line8 = draw.translate(line7, 0.0, 2.0 * p.turn_radius)
        line9 = draw.translate(line8, 0.0, 2.0 * p.turn_radius)
        line10 = draw.translate(line9, 0.0, 2.0 * p.turn_radius)
        line11 = draw.translate(line10, 0.0, 2.0 * p.turn_radius)
        line12 = draw.translate(line11, 0.0, 2.0 * p.turn_radius)
        line13 = draw.translate(line12, 0.0, 2.0 * p.turn_radius)
        line14 = draw.translate(line13, 0.0, 2.0 * p.turn_radius)
        # hard code the position for now; make it adjustable later
        line_last = draw.translate(
            line1, 0.5 * p.box_width - 0.0 * p.turn_radius - p.outer_space,
            28 * p.turn_radius)
        # draw the line exiting the rectangle
        final = draw.LineString([
            (0.0, p.perimeter_thickness + p.initial + 30.0 * p.turn_radius),
            (0.0,
             p.perimeter_thickness + p.initial + 30.0 * p.turn_radius + p.final)
        ])
        arc_last = draw.rotate(arc,
                               180,
                               origin=(0.0, p.perimeter_thickness + p.initial +
                                       1.5 * p.turn_radius))
        arc_last = draw.translate(arc_last, 0.0, p.initial + 26 * p.turn_radius)
        # repeat the left turns
        arc_left_2 = draw.translate(arc_left_1, 0.0, 4.0 * p.turn_radius)
        arc_left_3 = draw.translate(arc_left_2, 0.0, 4.0 * p.turn_radius)
        arc_left_4 = draw.translate(arc_left_3, 0.0, 4.0 * p.turn_radius)
        arc_left_5 = draw.translate(arc_left_4, 0.0, 4.0 * p.turn_radius)
        arc_left_6 = draw.translate(arc_left_5, 0.0, 4.0 * p.turn_radius)
        arc_left_7 = draw.translate(arc_left_6, 0.0, 4.0 * p.turn_radius)
        # repeat the right turns
        arc_right_2 = draw.translate(arc_right_1, 0.0, 4.0 * p.turn_radius)
        arc_right_3 = draw.translate(arc_right_2, 0.0, 4.0 * p.turn_radius)
        arc_right_4 = draw.translate(arc_right_3, 0.0, 4.0 * p.turn_radius)
        arc_right_5 = draw.translate(arc_right_4, 0.0, 4.0 * p.turn_radius)
        arc_right_6 = draw.translate(arc_right_5, 0.0, 4.0 * p.turn_radius)
        arc_right_7 = draw.translate(arc_right_6, 0.0, 4.0 * p.turn_radius)
        # Translate and rotate all shapes
        objects = [
            perimeter, initial, arc, arc_left_1, arc_left_2, arc_left_3,
            arc_left_4, arc_left_5, arc_left_6, arc_left_7, arc_right_1,
            arc_right_2, arc_right_2, arc_right_3, arc_right_4, arc_right_4,
            arc_right_5, arc_right_6, arc_right_7, arc_last, final, line1,
            line2, line3, line4, line5, line6, line7, line8, line9, line10,
            line11, line12, line13, line14, line_last
        ]
        # first translate so that the origin is at the middle of the loop
        objects = draw.translate(objects, 0.0, -0.5 * p.box_height)
        # now translate and rotate according to the values specified in the dictionary
        objects = draw.rotate(objects, p.orientation, origin=(0, 0))
        objects = draw.translate(objects, xoff=p.pos_x, yoff=p.pos_y)
        [
            perimeter, initial, arc, arc_left_1, arc_left_2, arc_left_3,
            arc_left_4, arc_left_5, arc_left_6, arc_left_7, arc_right_1,
            arc_right_2, arc_right_2, arc_right_3, arc_right_4, arc_right_4,
            arc_right_5, arc_right_6, arc_right_7, arc_last, final, line1,
            line2, line3, line4, line5, line6, line7, line8, line9, line10,
            line11, line12, line13, line14, line_last
        ] = objects
        # give polys names for qgeometry
        geom_perimeter = {'poly1': perimeter}
        geom_initial = {'poly2': initial}
        geom_arc = {'poly3': arc}
        geom_arc_left_1 = {'poly4': arc_left_1}
        geom_arc_left_2 = {'poly4': arc_left_2}
        geom_arc_left_3 = {'poly4': arc_left_3}
        geom_arc_left_4 = {'poly4': arc_left_4}
        geom_arc_left_5 = {'poly4': arc_left_5}
        geom_arc_left_6 = {'poly4': arc_left_6}
        geom_arc_left_7 = {'poly4': arc_left_7}
        geom_arc_right_1 = {'poly4': arc_right_1}
        geom_arc_right_2 = {'poly4': arc_right_2}
        geom_arc_right_3 = {'poly4': arc_right_3}
        geom_arc_right_4 = {'poly4': arc_right_4}
        geom_arc_right_5 = {'poly4': arc_right_5}
        geom_arc_right_6 = {'poly4': arc_right_6}
        geom_arc_right_7 = {'poly4': arc_right_7}
        geom_arc_last = {'poly4': arc_last}
        geom_final = {'poly': final}
        geom_line1 = {'poly5': line1}
        geom_line2 = {'poly6': line2}
        geom_line3 = {'poly7': line3}
        geom_line4 = {'poly8': line4}
        geom_line5 = {'poly9': line5}
        geom_line6 = {'poly10': line6}
        geom_line7 = {'poly11': line7}
        geom_line8 = {'poly11': line8}
        geom_line9 = {'poly9': line9}
        geom_line10 = {'poly10': line10}
        geom_line11 = {'poly11': line11}
        geom_line12 = {'poly12': line12}
        geom_line13 = {'poly13': line13}
        geom_line14 = {'poly14': line14}
        geom_line_last = {'poly14': line_last}
        # add to qgeometry
        self.add_qgeometry('poly',
                           geom_perimeter,
                           layer=p.layer,
                           subtract=False)
        self.add_qgeometry('path',
                           geom_initial,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_left_1,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_left_2,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_left_3,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_left_4,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_left_5,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_left_6,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_left_7,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_right_1,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_right_2,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_right_3,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_right_4,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_right_5,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_right_6,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_right_7,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_arc_last,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line1,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line2,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line3,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line4,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line5,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line6,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line7,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line8,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line9,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line10,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line11,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line12,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line13,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line14,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_line_last,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)
        self.add_qgeometry('path',
                           geom_final,
                           layer=p.layer,
                           subtract=False,
                           width=p.res_width)

        ##########################
        # Add Qpin connections
        # define a function that both rotates and translates the qpin coordinates
        def qpin_rotate_translate(x):
            y = list(x)
            z = [0.0, 0.0]
            z[0] = y[0] * cos(p.orientation * 3.14159 / 180) - y[1] * sin(
                p.orientation * 3.14159 / 180)
            z[1] = y[0] * sin(p.orientation * 3.14159 / 180) + y[1] * cos(
                p.orientation * 3.14159 / 180)
            z[0] = z[0] + p.pos_x
            z[1] = z[1] + p.pos_y
            x = (z[0], z[1])
            return x

        # East pin
        qp1a = (0.5 * p.box_width - p.perimeter_thickness, 0.0)
        qp1b = (0.5 * p.box_width, 0.0)
        qp1a = qpin_rotate_translate(qp1a)
        qp1b = qpin_rotate_translate(qp1b)
        self.add_pin('pin_east',
                     points=np.array([qp1a, qp1b]),
                     width=0.01,
                     input_as_norm=True)
        # North-East pin
        qpa_ne = (0.5 * p.box_width - p.perimeter_thickness,
                  0.5 * p.box_height - p.perimeter_thickness)
        qpb_ne = (0.5 * p.box_width, 0.5 * p.box_height)
        qpa_ne = qpin_rotate_translate(qpa_ne)
        qpb_ne = qpin_rotate_translate(qpb_ne)
        self.add_pin('pin_ne',
                     points=np.array([qpa_ne, qpb_ne]),
                     width=0.01,
                     input_as_norm=True)
        # South-East pin
        qpa_se = (0.5 * p.box_width - p.perimeter_thickness,
                  -0.5 * p.box_height + p.perimeter_thickness)
        qpb_se = (0.5 * p.box_width, -0.5 * p.box_height)
        qpa_se = qpin_rotate_translate(qpa_se)
        qpb_se = qpin_rotate_translate(qpb_se)
        self.add_pin('pin_se',
                     points=np.array([qpa_se, qpb_se]),
                     width=0.01,
                     input_as_norm=True)
        # North-West pin
        qpa_nw = (-0.5 * p.box_width + p.perimeter_thickness,
                  0.5 * p.box_height - p.perimeter_thickness)
        qpb_nw = (-0.5 * p.box_width, 0.5 * p.box_height)
        qpa_nw = qpin_rotate_translate(qpa_nw)
        qpb_nw = qpin_rotate_translate(qpb_nw)
        self.add_pin('pin_nw',
                     points=np.array([qpa_nw, qpb_nw]),
                     width=0.01,
                     input_as_norm=True)
        # West pin
        qp2a = (-0.5 * p.box_width + p.perimeter_thickness, 0.0)
        qp2b = (-0.5 * p.box_width, 0.0)
        qp2a = qpin_rotate_translate(qp2a)
        qp2b = qpin_rotate_translate(qp2b)
        self.add_pin('pin_west',
                     points=np.array([qp2a, qp2b]),
                     width=0.01,
                     input_as_norm=True)
        # South-West pin
        qpa_sw = (-0.5 * p.box_width + p.perimeter_thickness,
                  -0.5 * p.box_height + p.perimeter_thickness)
        qpb_sw = (-0.5 * p.box_width, -0.5 * p.box_height)
        qpa_sw = qpin_rotate_translate(qpa_sw)
        qpb_sw = qpin_rotate_translate(qpb_sw)
        self.add_pin('pin_sw',
                     points=np.array([qpa_sw, qpb_sw]),
                     width=0.01,
                     input_as_norm=True)
        # South pin
        qpa_s = (0.0, -0.5 * p.box_height + p.perimeter_thickness)
        qpb_s = (0.0, -0.5 * p.box_height)
        qpa_s = qpin_rotate_translate(qpa_s)
        qpb_s = qpin_rotate_translate(qpb_s)
        self.add_pin('pin_s',
                     points=np.array([qpa_s, qpb_s]),
                     width=0.01,
                     input_as_norm=True)
        # north pin
        qpa_n = (0.0, -0.5 * p.box_height + p.perimeter_thickness + p.initial +
                 30.0 * p.turn_radius)
        qpb_n = (0.0, -0.5 * p.box_height + p.perimeter_thickness + p.initial +
                 30.0 * p.turn_radius + p.final)
        qpa_n = qpin_rotate_translate(qpa_n)
        qpb_n = qpin_rotate_translate(qpb_n)
        self.add_pin('pin_n',
                     points=np.array([qpa_n, qpb_n]),
                     width=0.01,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Concentric Transmon.
"""

from math import sin, cos
import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import BaseQubit


class TransmonConcentric(BaseQubit):
    """The base `TrasmonConcentric` class .

    Inherits `BaseQubit` class.

    Metal transmon object consisting of a circle surrounding by a concentric
    ring. There are two Josephson Junction connecting the circle to the ring;
    one at the south end and one at the north end. There is a readout resonator.

    .. image::
        TransmonConcentric.png

    .. meta::
        :description: Transmon Concentric

    BaseQubit Default Options:
        * connection_pads: empty Dict -- the dictionary which contains all active
          connection lines for the qubit.
        * _default_connection_pads: empty Dict -- the default values for the
          (if any) connection lines of the qubit.

    Default Options:
        * width: '1000um' -- Width of transmon pocket
        * height: '1000um' -- Height of transmon pocket
        * rad_o: '170um' -- Outer radius
        * rad_i: '115um' -- Inner radius
        * gap: '35um' -- Radius of gap between two pads
        * jj_w: '10um' -- Josephson Junction width
        * res_s: '100um' -- Space between top electrode and readout resonator
        * res_ext: '100um' -- Extension of readout resonator in x-direction
          beyond midpoint of transmon
        * fbl_rad: '100um' -- Radius of the flux bias line loop
        * fbl_sp: '100um' -- Spacing between metal pad and flux bias loop
        * fbl_gap: '80um' -- Space between parallel lines of the flux bias loop
        * fbl_ext: '300um' -- Run length of flux bias line between circular
          loop and edge of pocket
        * pocket_w: '1500um' -- Transmon pocket width
        * pocket_h: '1000um' -- Transmon pocket height
        * cpw_width: '10.0um' -- Width of the readout resonator and flux bias line
        * layer: '1' -- the design layer where the component will go
    """

    # default drawing options
    default_options = Dict(
        width='1000um',  # width of transmon pocket
        height='1000um',  # height of transmon pocket
        rad_o='170um',  # outer radius
        rad_i='115um',  # inner radius
        gap='35um',  # radius of gap between two pads
        jj_w='10um',  # Josephson Junction width
        res_s='100um',  # space between top electrode and readout resonator
        res_ext=
        '100um',  # extension of readout resonator in x-direction beyond midpoint of transmon
        fbl_rad='100um',  # radius of the flux bias line loop
        fbl_sp='100um',  # spacing between metal pad and flux bias loop
        fbl_gap='80um',  # space between parallel lines of the flux bias loop
        fbl_ext=
        '300um',  # run length of flux bias line between circular loop and edge of pocket
        pocket_w='1500um',  # transmon pocket width
        pocket_h='1000um',  # transmon pocket height
        cpw_width='10.0um',  # width of the readout resonator and flux bias line
        inductor_width='5.0um',  # width of the Josephson Junctions
        layer='1'  # design layer
    )
    """Default drawing options"""

    TOOLTIP = """The base `TrasmonConcentric` class."""

    def make(self):
        """Convert self.options into QGeometry."""

        p = self.parse_options()  # Parse the string options into numbers

        # draw the concentric pad regions
        outer_pad = draw.Point(0, 0).buffer(p.rad_o)
        space = draw.Point(0, 0).buffer((p.gap + p.rad_i))
        outer_pad = draw.subtract(outer_pad, space)
        inner_pad = draw.Point(0, 0).buffer(p.rad_i)
        #gap = draw.subtract(space, inner_pad)
        #pads = draw.union(outer_pad, inner_pad)

        # draw the top Josephson Junction
        jj_t = draw.LineString([(0.0, p.rad_i), (0.0, p.rad_i + p.gap)])

        # draw the bottom Josephson Junction
        jj_b = draw.LineString([(0.0, -1.0 * p.rad_i),
                                (0.0, -1.0 * p.rad_i - 1.0 * p.gap)])

        # draw the readout resonator
        qp1a = (-0.5 * p.pocket_w, p.rad_o + p.res_s
               )  # the first (x,y) coordinate is qpin #1
        qp1b = (p.res_ext, p.rad_o + p.res_s
               )  # the second (x,y) coordinate is qpin #1
        rr = draw.LineString([qp1a, qp1b])

        # draw the flux bias line
        a = (0.5 * p.pocket_w, -0.5 * p.fbl_gap)
        b = (0.5 * p.pocket_w - p.fbl_ext, -0.5 * p.fbl_gap)
        c = (p.rad_o + p.fbl_sp + p.fbl_rad, -1.0 * p.fbl_rad)
        d = (p.rad_o + p.fbl_sp + 0.2929 * p.fbl_rad, 0.0 - 0.7071 * p.fbl_rad)
        e = (p.rad_o + p.fbl_sp, 0.0)
        f = (p.rad_o + p.fbl_sp + 0.2929 * p.fbl_rad, 0.0 + 0.7071 * p.fbl_rad)
        g = (p.rad_o + p.fbl_sp + p.fbl_rad, p.fbl_rad)
        h = (0.5 * p.pocket_w - p.fbl_ext, 0.5 * p.fbl_gap)
        i = (0.5 * p.pocket_w, 0.5 * p.fbl_gap)
        fbl = draw.LineString([a, b, c, d, e, f, g, h, i])

        # draw the transmon pocket bounding box
        pocket = draw.rectangle(p.pocket_w, p.pocket_h)

        # Translate and rotate all shapes
        objects = [outer_pad, inner_pad, jj_t, jj_b, pocket, rr, fbl]
        objects = draw.rotate(objects, p.orientation, origin=(0, 0))
        objects = draw.translate(objects, xoff=p.pos_x, yoff=p.pos_y)
        [outer_pad, inner_pad, jj_t, jj_b, pocket, rr, fbl] = objects

        # define a function that both rotates and translates the qpin coordinates
        def qpin_rotate_translate(x):
            y = list(x)
            z = [0.0, 0.0]
            z[0] = y[0] * cos(p.orientation * 3.14159 / 180) - y[1] * sin(
                p.orientation * 3.14159 / 180)
            z[1] = y[0] * sin(p.orientation * 3.14159 / 180) + y[1] * cos(
                p.orientation * 3.14159 / 180)
            z[0] = z[0] + p.pos_x
            z[1] = z[1] + p.pos_y
            x = (z[0], z[1])
            return x

        # rotate and translate the qpin coordinates
        qp1a = qpin_rotate_translate(qp1a)
        qp1b = qpin_rotate_translate(qp1b)
        a = qpin_rotate_translate(a)
        b = qpin_rotate_translate(b)
        h = qpin_rotate_translate(h)
        i = qpin_rotate_translate(i)

        ##############################################################

        # Use the geometry to create Metal QGeometry
        geom_rr = {'path1': rr}
        geom_fbl = {'path2': fbl}
        geom_outer = {'poly1': outer_pad}
        geom_inner = {'poly2': inner_pad}
        geom_jjt = {'poly4': jj_t}
        geom_jjb = {'poly5': jj_b}
        geom_pocket = {'poly6': pocket}

        self.add_qgeometry('path',
                           geom_rr,
                           layer=p.layer,
                           subtract=False,
                           width=p.cpw_width)
        self.add_qgeometry('path',
                           geom_fbl,
                           layer=p.layer,
                           subtract=False,
                           width=p.cpw_width)
        self.add_qgeometry('poly', geom_outer, layer=p.layer, subtract=False)
        self.add_qgeometry('poly', geom_inner, layer=p.layer, subtract=False)
        self.add_qgeometry('junction',
                           geom_jjt,
                           layer=p.layer,
                           subtract=False,
                           width=p.inductor_width)
        self.add_qgeometry('junction',
                           geom_jjb,
                           layer=p.layer,
                           subtract=False,
                           width=p.inductor_width)
        self.add_qgeometry('poly', geom_pocket, layer=1, subtract=True)

        ###########################################################################

        # Add Qpin connections
        self.add_pin('pin1',
                     points=np.array([qp1b, qp1a]),
                     width=0.01,
                     input_as_norm=True)
        self.add_pin('pin2',
                     points=np.array([b, a]),
                     width=0.01,
                     input_as_norm=True)
        self.add_pin('pin3',
                     points=np.array([h, i]),
                     width=0.01,
                     input_as_norm=True)
